# -----------------------------------------------------------------------------
#   Copyright (C): OpenGATE Collaboration
#   This software is distributed under the terms
#   of the GNU Lesser General  Public Licence (LGPL)
#   See LICENSE.md for further details
# -----------------------------------------------------------------------------

"""

This module provides a function to convert image from one type to another:

"""

# -----------------------------------------------------------------------------
#   Copyright (C): OpenGATE Collaboration
#   This software is distributed under the terms
#   of the GNU Lesser General  Public Licence (LGPL)
#   See LICENSE.md for further details
# -----------------------------------------------------------------------------


import itk
import pydicom
from pydicom.tag import Tag
import gatetools as gt
import numpy as np
import logging
logger=logging.getLogger(__name__)

def convertRadiusToPx(radius, spacing):
    newRadius = []
    for r, s in zip(radius, spacing):
        newRadius.append(int(r/s))
    return(newRadius)


def morpho_math(inputImage, morphotype, radius):
    """
    Compute morpho operation
    """
    ImageType = type(inputImage)

    #Set the kernel element
    StructuringElementType = itk.FlatStructuringElement[inputImage.GetImageDimension()]
    structuringElement = StructuringElementType.Ball(radius)
    structuringElement.SetRadiusIsParametric(True)

    #Choose the morphological operation
    if morphotype == 0:
        MorphoFilterType = itk.BinaryDilateImageFilter[ImageType, ImageType, StructuringElementType]
    elif morphotype == 1:
        MorphoFilterType = itk.BinaryErodeImageFilter[ImageType, ImageType, StructuringElementType]
    elif morphotype == 2:
        MorphoFilterType = itk.BinaryMorphologicalClosingImageFilter[ImageType, ImageType, StructuringElementType]
    elif morphotype == 3:
        MorphoFilterType = itk.BinaryMorphologicalOpeningImageFilter[ImageType, ImageType, StructuringElementType]
    else:
        print("Error: choose a correct morpho type")
        return None

    #Compute the operation
    morphoFilter = MorphoFilterType.New()
    morphoFilter.SetInput(inputImage)
    morphoFilter.SetKernel(structuringElement)
    morphoFilter.SetForegroundValue(1)
    morphoFilter.Update()
    output = morphoFilter.GetOutput()
    output.CopyInformation(inputImage)
    return (output)



#####################################################################################
import unittest
import hashlib
import os
import tempfile
import shutil
import wget
from .logging_conf import LoggedTestCase

def createImage():
        Dimension = 3
        PixelType = itk.ctype('unsigned char')
        ImageType = itk.Image[PixelType, Dimension]

        image = ImageType.New()
        start = itk.Index[Dimension]()
        start[0] = 0
        start[1] = 0
        start[2] = 0
        size = itk.Size[Dimension]()
        size[0] = 200
        size[1] = 200
        size[2] = 200
        region = itk.ImageRegion[Dimension]()
        region.SetSize(size)
        region.SetIndex(start)
        image.SetRegions(region)
        image.Allocate()
        image.FillBuffer(0)
        npView = itk.array_from_image(image)
        npView[10:52, 42:192, 124:147] =1
        npView[26:30, 42:149, 132:138] =0
        npView[26:30, 151:192, 132:138] =0
        image = itk.image_from_array(npView)
        return image

class Test_Morpho_Math(LoggedTestCase):
    def test_morpho_math_dilatation(self):
        image = createImage()
        output = morpho_math(image, 0, [2, 3, 4])
        outputStats = gt.imageStatistics(input=output)
        self.assertTrue(outputStats["sum"] == 208836)
    def test_morpho_math_dilatation_mm(self):
        image = createImage()
        image.SetSpacing([2, 3, 4])
        radius = [2, 3, 4]
        newRadius = convertRadiusToPx(radius, image.GetSpacing())
        output = morpho_math(image, 0, newRadius)
        outputStats = gt.imageStatistics(input=output)
        self.assertTrue(outputStats["sum"] == 166008)
    def test_morpho_math_erosion(self):
        image = createImage()
        output = morpho_math(image, 1, [2, 3, 4])
        outputStats = gt.imageStatistics(input=output)
        self.assertTrue(outputStats["sum"] == 76904)
    def test_morpho_math_closing(self):
        image = createImage()
        output = morpho_math(image, 2, [2, 3, 4])
        outputStats = gt.imageStatistics(input=output)
        self.assertTrue(outputStats["sum"] == 144900)
    def test_morpho_math_opening(self):
        image = createImage()
        output = morpho_math(image, 3, [2, 3, 4])
        outputStats = gt.imageStatistics(input=output)
        self.assertTrue(outputStats["sum"] == 139576)




# -----------------------------------------------------------------------------
#   Copyright (C): OpenGATE Collaboration
#   This software is distributed under the terms
#   of the GNU Lesser General  Public Licence (LGPL)
#   See LICENSE.md for further details
# -----------------------------------------------------------------------------

"""
This module contains a function named `dicom_rt_pbs_plan_to_gate_conversion`,
which can be used to convert DICOM radiotherapy plan files for pencil beam
scanning to the text format that Gate uses to read in the spot specifications.
"""

import numpy as np
import logging
logger=logging.getLogger(__name__)

def dicom_rt_pbs_plan_to_gate_conversion(dcm_input,txt_output,allow0=False,verbose=False):
    """
    Function to create GATE plan description corresponding to DICOM plan file.

    * :param dcm_input: string with the file path of the input DICOM plan file
    * :param txt_output: string with the file path of the output Gate PBS spot specification text file (suffix .txt or .dat)
    """
    planname = _check_output_filename(txt_output,verbose)
    nspots_written = 0
    nlayers_written = 0
    nspots_ignored = 0
    nlayers_ignored = 0
    rp,mswtots,beamnrs,iso_Cs,g_angles,p_angles = _read_and_check_dicom_plan_file(dcm_input,verbose)
    filehandle=open(txt_output,"w")
    # FILE HEADER
    filehandle.write(_file_header.format(name=planname,nfields=len(rp.IonBeamSequence)))
    for beamnr in beamnrs:
        filehandle.write("###FieldsID\n{}\n".format(beamnr))
    filehandle.write("#TotalMetersetWeightOfAllFields\n{0:f}\n\n".format(sum(mswtots)))
    # stupidity check...
    assert(len(rp.IonBeamSequence)== len(beamnrs))
    assert(len(rp.IonBeamSequence)== len(mswtots))
    assert(len(rp.IonBeamSequence)== len(iso_Cs))
    assert(len(rp.IonBeamSequence)== len(g_angles))
    assert(len(rp.IonBeamSequence)== len(p_angles))
    for ion_beam,beamnr,mswtot,iso_C,g_angle,p_angle in zip(rp.IonBeamSequence,
                                                           beamnrs,
                                                           mswtots,
                                                           iso_Cs,
                                                           g_angles,
                                                           p_angles):
        nspots_list=list()
        mask_list=list()
        weights_list=list()
        for icp in ion_beam.IonControlPointSequence:
            nspots_nominal = int(icp.NumberOfScanSpotPositions)
            if nspots_nominal == 1:
                w_all = np.array([float(icp.ScanSpotMetersetWeights)])
            else:
                w_all = np.array([float(w) for w in icp.ScanSpotMetersetWeights])
            mask = (w_all>=0.) if allow0 else (w_all>0.)
            nspots_list.append(np.sum(mask))
            mask_list.append(mask)
            weights_list.append(w_all)

        filehandle.write(_beam_header.format(fid=beamnr,
                                             mswtot=mswtot,
                                             g_angle=g_angle,
                                             p_angle=p_angle,
                                             isox=iso_C[0],
                                             isoy=iso_C[1],
                                             isoz=iso_C[2],
                                             ncp=np.sum(np.array(nspots_list)>0)))
        msw_cumsum = 0.
        cpi=0
        for icp,nspots,mask,w_all in zip(ion_beam.IonControlPointSequence,nspots_list,mask_list,weights_list):
            nspots_nominal = int(icp.NumberOfScanSpotPositions)
            if nspots == 0:
                if allow0:
                    logger.warning("this should not happen, nspots_nominal={}",format(nspots_nominal))
                nlayers_ignored += 1
                nspots_ignored += nspots_nominal
                continue
            xy = np.array([float(pos) for pos in icp.ScanSpotPositionMap]).reshape(nspots_nominal,2)
            nspots_ignored += nspots_nominal-nspots
            w=w_all[mask]
            x=xy[mask,0]
            y=xy[mask,1]
            msw=np.sum(w_all[mask])
            energy = float(icp.NominalBeamEnergy)
            # Maybe we should emit noisy warnings if the tune ID is missing?
            tuneID = "0.0" if not "ScanSpotTuneID" in icp else str(icp.ScanSpotTuneID)
            # TODO: what if all spot weights in this ICP are zero?
            filehandle.write(_layer_header.format(cpi=cpi,stid=tuneID,mswtot=msw_cumsum,energy=energy,nspots=nspots))
            for spotx,spoty,spotw in zip(x,y,w):
                filehandle.write("{0:g} {1:g} {2:g}\n".format(spotx,spoty,spotw))
            nspots_written += nspots
            nlayers_written += 1
            msw_cumsum+=msw
            cpi+=1
    filehandle.close()
    if verbose:
        logger.info("Converted DICOM file {} to Gate PBS spot specification text file {}".format(dcm_input,txt_output))
        logger.info("# beams = {}".format(len(rp.IonBeamSequence)))
        logger.info("# control points written = {}".format(nlayers_written))
        logger.info("# spots written = {}".format(nspots_written))
        if not allow0:
            # number of beams ignored, because msw<=0 for all spots? :-)
            logger.info("# control points ignored (because msw<=0 for all spots) = {}".format(nlayers_ignored))
            logger.info("# spots ignored (because msw<=0) = {}".format(nspots_ignored))

###############################################################################
# IMPLEMENTATION DETAILS                                                      #
###############################################################################

import os, re
import pydicom

def _check_rp_dicom_file(rp_filepath,verbose=False):
    """
    Auxiliary implementation function.

    Try to read the DICOM file and perform some paranoid check of the essential DICOM attributes.
    """
    # if the input file path is not readable as a DICOM file, then pydicom will throw an appropriate exception
    rp = pydicom.dcmread(rp_filepath)
    for attr in [ 'SOPClassUID', "IonBeamSequence" ] :
        if attr not in rp:
            raise IOError("bad DICOM file {},\nmissing '{}'".format(rp_filepath,attr))
    if rp.SOPClassUID.name != 'RT Ion Plan Storage':
        raise IOError("bad plan file {},\nwrong SOPClassUID: {}='{}',\nexpecting an 'RT Ion Plan Storage' file instead.".format(rp_filepath,rp.SOPClassUID,rp.SOPClassUID.name))
    n_ion_beams=len(rp.IonBeamSequence)
    ion_beams=rp.IonBeamSequence
    for ion_beam in ion_beams:
        for attr in [ 'BeamNumber', "IonControlPointSequence"]:
            if attr not in ion_beam:
                raise IOError("bad DICOM file {},\nmissing '{}' in ion beam".format(rp_filepath,attr))
        for cpi,icp in enumerate(ion_beam.IonControlPointSequence):
            for attr in [ 'NumberOfScanSpotPositions', "ScanSpotPositionMap", "ScanSpotMetersetWeights"]:
                if attr not in icp:
                    raise IOError("bad DICOM file {},\nmissing '{}' in {}th ion control point".format(rp_filepath,attr,cpi))
    if verbose:
        logger.info("Input DICOM file seems to be a legit 'RT Ion Plan Storage' file.")
    return rp

def _get_beam_numbers(rp,verbose=False):
    """
    Auxiliary implementation function.

    Extract beam numbers from the DICOM plan dataset, and create better ones in
    case the stored ones are useless for some reason.

    Most treatment planning systems will assign a non-negative unique beam
    number to each ion beam, but some TPS (ones used to generate artificial
    plans for testing and commissioning purposes) are negligent in that regard.
    Since Gate needs the beam numbers in order to allow the user to select
    which beams to simulate (or conversely, which beams to ignore), this script
    will create fake but usable beam numbers in such cases.
    """
    number_list = list()
    input_beam_numbers_are_ok = True
    n_ion_beams=len(rp.IonBeamSequence)
    for ion_beam in rp.IonBeamSequence:
        if input_beam_numbers_are_ok:
            nr = int(ion_beam.BeamNumber)
            if nr < 0:
                input_beam_numbers_are_ok = False
                if verbose:
                    logger.info("CORRUPT INPUT: found a negative beam number {}.".format(nr))
            elif nr in number_list:
                input_beam_numbers_are_ok = False
                if verbose:
                    logger.info("CORRUPT INPUT: found same beam number {} for multiple beams.".format(nr))
            else:
                # still good, keep fingers crossed...
                number_list.append(nr)
    if not input_beam_numbers_are_ok:
        if verbose:
            logger.warning("will use simple enumeration of beams instead of the (apparently corrupt) dicom beam numbers.")
        number_list = np.arange(1,n_ion_beams+1).tolist()
    return number_list

def _get_mswtot_list(rp,verbose=False):
    """
    Auxiliary implementation function.

    Retrieve the total weight of all spots for one "control point" (or "layer",
    or "energy").  TODO: it could be useful to enable a conversion function
    here, e.g. to convert from "monitor units" to "number of protons". However,
    this can also be taken care of in Gate itself, via the beam calibration
    polynomial in the "source properties file".
    """
    mswtot_list = list()
    for ion_beam in rp.IonBeamSequence:
        mswtot=0.
        for icp in ion_beam.IonControlPointSequence:
            nspot =int(icp.NumberOfScanSpotPositions)
            if nspot == 1:
                mswtot += float(icp.ScanSpotMetersetWeights)
            else:
                # Weights should be non-negative, but let's be paranoid.
                mswtot += np.sum(np.array([float(w) for w in icp.ScanSpotMetersetWeights if w>0.]))
        mswtot_list.append(mswtot)
    return mswtot_list

def _get_angles_and_isoCs(rp,verbose):
    """
    Auxiliary implementation function.

    For each beam, retrieve the gantry angle, patient support angle and
    isocenter, if available.  (Most TPS will of course specify this info, but
    some medical physicists use an in-house TPS for generating special plans
    for beam verification and commissioning, with incomplete planning
    information.) For some unclear reasons, the angles and isocenter
    coordinates are not stored as attributes of the ion beam, but rather as
    attributes to the first "ion control point".
    """
    gantry_angle_list = list()
    patient_angle_list = list()
    iso_center_list = list()
    n_ion_beams=len(rp.IonBeamSequence)
    dubious = False
    for i,ion_beam in enumerate(rp.IonBeamSequence):
        # each of these quantities may be missing
        beamname = str(i) if not hasattr(ion_beam,"BeamName") else str(ion_beam.BeamName)
        icp0 = ion_beam.IonControlPointSequence[0]
        # check isocenter
        if "IsocenterPosition" in icp0 and len(icp0.IsocenterPosition)==3:
            iso_center_list.append(np.array([float(icp0.IsocenterPosition[j]) for j in range(3)]))
        else:
            logger.warning("absent/corrupted isocenter for beam '{}'; assuming [0,0,0] for now, please fix this manually.".format(beamname))
            iso_center_list.append(np.zeros(3,dtype=float))
            dubious = True
        # check gantry angle
        if "GantryAngle" in icp0:
            gantry_angle_list.append(float(icp0.GantryAngle))
        else:
            logger.warning("no gantry angle specified for beam '{}' in treatment plan; assuming 0. for now, please fix this manually.".format(beamname))
            gantry_angle_list.append(0.)
            dubious = True
        # check couch angle
        if "PatientSupportAngle" in icp0:
            patient_angle_list.append(float(icp0.PatientSupportAngle))
        else:
            logger.warning("no patient support angle specified for beam '{}' in treatment plan; assuming 0. for now, please fix this manually.".format(beamname))
            patient_angle_list.append(0.)
            dubious = True
    if verbose and not dubious:
         logger.info("patient/gantry angles and isocenters all seem fine.")
    return gantry_angle_list, patient_angle_list, iso_center_list

def _read_and_check_dicom_plan_file(rp_filepath,verbose=False):
    """
    Auxiliary implementation function.

    The existence of a DICOM 'standard' does not mean that all 'DICOM plan
    files' look alike, unfortunately; every TPS has its own dialect, and some medical
    physicists use their own hobby TPS to create "DICOM plan files" that are lacking
    the most basic ingredients. We need to check our assumptions based on the
    TPS plan files that we have had access to, and define workarounds for the
    problematic "plan files".
    """
    # Crude checks of DICOM file structure.
    rp = _check_rp_dicom_file(rp_filepath,verbose)
    # Get mswtot of each beam.
    mswtot_list = _get_mswtot_list(rp,verbose)
    # Get 'number' of each beam.
    # (Name would more more useful, but Gate uses the number in its interface for "allowing" and "disallowing" beams ("fields").)
    number_list = _get_beam_numbers(rp,verbose)
    # Get the things that *should* be attributes of an "ion beam" object but which are buried in "control point number 0".
    gantry_angles,patient_angles,isoCs = _get_angles_and_isoCs(rp,verbose)

    return rp, mswtot_list, number_list, isoCs, gantry_angles, patient_angles

def _check_output_filename(txt_output,verbose):
    """
    Let's assume that the user chose an informative name for the spot file,
    and use that for the plan name. Gate does not actually use that name,
    so it's not so important. But it *is* important to make sure that it is
    a single word (without spaces or other non-alphanumerical characters).
    """
    if txt_output[-4:].lower() != ".txt" and txt_output[-4:].lower() != ".dat":
        raise IOError("Output file name should have a '.txt' or '.dat' suffix.")
    base_txt_output = os.path.basename(txt_output)
    if len(base_txt_output)<5:
        raise IOError("Output file name should have at least one charachter before the txt/dat suffix.")
    if os.path.exists(txt_output) and verbose:
        logger.warning("going to overwrite existing file {}".format(txt_output))
    badchars=re.compile("[^a-zA-Z0-9_]")
    planname = re.sub(badchars,"_",base_txt_output[:-4])
    if verbose:
        logger.info("using plan name '{}'".format(planname))
    return planname



_file_header="""#TREATMENT-PLAN-DESCRIPTION
#PlanName
{name:s}
#NumberOfFractions
1
##FractionID
1
##NumberOfFields
{nfields:d}
"""

_beam_header="""#FIELD-DESCRIPTION
###FieldID
{fid}
###FinalCumulativeMeterSetWeight
{mswtot:g}
###GantryAngle (in degrees)
{g_angle:g}
###PatientSupportAngle
{p_angle}
###IsocenterPosition
{isox:g} {isoy:g} {isoz:g}
###NumberOfControlPoints
{ncp:d}
#SPOTS-DESCRIPTION
"""

_layer_header="""####ControlPointIndex
{cpi:d}
####SpotTuneID
{stid:s}
####CumulativeMetersetWeight
{mswtot:g}
####Energy (MeV)
{energy:g}
####NbOfScannedSpots
{nspots:g}
####X Y Weight (spot position at isocenter in mm, with weight in MU (default) or number of protons "setSpotIntensityAsNbProtons true")
"""

################################################################################
# UNIT TESTS                                                                   #
################################################################################

import unittest
import sys
import datetime

class _tmp_test_plan_writer:
    def __init__(self,fname,spotspecs=None,uid=None,verbose=False):
        """
        Create a fake treatment plan DICOM file for unit testing.  The spot
        specs are a list of dictionaries, one dictionary per beam.  Each beam
        dictionary contains optional entries for gantry angle, patient support
        angle and isocenter position and an obligatory list of "controlpoints".

        A file with filename `fname` should not yet exist. A DICOM file with
        name `fname` will be written by _test_plan_creator, and also be deleted
        when the creator goes out of scope.
        """
        assert(sys.version_info.major == 3)
        self.verbose=verbose
        assert(not os.path.exists(fname))
        self.dcm_filename=fname
        classUID='1.2.840.10008.5.1.4.1.1.481.8' if uid is None else uid
        instanceUID=pydicom.uid.generate_uid()

        # File meta info data elements
        file_meta = pydicom.Dataset()
        file_meta.MediaStorageSOPClassUID = str(classUID)
        file_meta.MediaStorageSOPInstanceUID = str(instanceUID)
        file_meta.TransferSyntaxUID = pydicom.uid.ImplicitVRLittleEndian
        #FIXME: we probably need to apply for an official UID here
        file_meta.ImplementationClassUID = '1.2.826.0.1.3680043.1.2.100.6.40.0.76'
        file_meta.ImplementationVersionName = 'DicomObjects.NET'

        self.ds=pydicom.FileDataset(self.dcm_filename,{},file_meta=file_meta,preamble=b"\0"*128)
        self.ds.SOPClassUID=classUID
        self.ds.SOPInstanceUID=instanceUID
        self.ds.PatientName="unit test"
        self.ds.PatientID="42"
        #print("trying to fix 'is_little_endian'")

        if spotspecs:
            self.ds.IonBeamSequence=pydicom.Sequence()
            self.ds.NumberOfBeams = len(spotspecs)
            for beamdata in spotspecs:
                beam=pydicom.Dataset()
                beam.BeamName = beamdata.get('Nm','noname')
                beam.BeamNumber = beamdata.get('Nr',-1)
                beam.NumberOfIonControlPoints=len(beamdata["controlpoints"])
                beam.IonControlPointSequence=pydicom.Sequence()
                for i,icpdata in enumerate(beamdata["controlpoints"]):
                   icp=pydicom.Dataset()
                   if i==0:
                       if 'G' in beamdata:
                           icp.GantryAngle = beamdata['G']
                       if 'P' in beamdata:
                           icp.PatientSupportAngle = beamdata['P']
                       if "I" in beamdata:
                           icp.IsocenterPosition = list([str(v) for v in beamdata.get('I')])
                       else:
                           icp.IsocenterPosition = ""
                   icp.NominalBeamEnergy = icpdata.get('E',100.)
                   if 'T' in icpdata:
                       icp.ScanSpotTuneID = icpdata['T']
                   #unfortunately, this does not work
                   icp.ScanSpotMetersetWeights = [float(v) for v in icpdata['SSW']]
                   icp.ScanSpotPositionMap = [float(v) for v in np.array(icpdata['SSP']).flatten()]
                   icp.NumberOfScanSpotPositions = int(icpdata.get('NSSP',len(icp.ScanSpotPositionMap)/2))
                   beam.IonControlPointSequence.append(icp)
                self.ds.IonBeamSequence.append(beam)
        else:
            self.ds.NumberOfBeams = 0
        if verbose:
            logger.info("number of beams is {}".format(self.ds.NumberOfBeams))
        dt = datetime.datetime.now()
        self.ds.ContentDate = dt.strftime('%Y%m%d')
        timeStr = dt.strftime('%H%M%S.%f')  # long format with micro seconds
        self.ds.ContentTime = timeStr
        self.ds.is_little_endian = True
        self.ds.is_implicit_VR = True
        self.ds.fix_meta_info()
        self.ds.save_as(self.dcm_filename)
        if self.verbose:
            logger.info("wrote dicom file {}".format(self.dcm_filename))
    def __del__(self):
        os.remove(self.dcm_filename)
        if self.verbose:
            logger.info("deleted dicom file {}".format(self.dcm_filename))
        del self.ds

class test_small_normal_plan(unittest.TestCase):
    """
    Normal plan with three properly numbered beams and all data needed by Gate for a simulation.
    """
    def setUp(self):
        beam4 = dict(Nm="foo", Nr=4, G=45., P=225., I=(120.,130.,140.),
                     controlpoints=[dict(SSW=[0.42],E=100.,T="3.0",
                                         SSP=[(2.,1)])])
        beam5 = dict(Nm="bar", Nr=5, G=30., P=45., I=(100.,150.,-140.),
                     controlpoints=[dict(SSW=[0.,0.4,0.,0.6],E=100.,T="3.0",
                                         SSP=[(2.,1.),(2.,1),(3.,4.),(3.,4.)])])
        beam6 = dict(Nm="baz", Nr=6, G=60., P=30., I=(-10.,-150.,40.),
                     controlpoints=[dict(SSW=[0.4,0.6,0.8],E=70.,T="3.1",
                                         SSP=[(2.,1),(3.,4.),(6.,5.)]),
                                    dict(SSW=[0.4,0.5],E=120.,T="3.1",
                                         SSP=[(7.,8.),(3.,4.)])])
        self.testbeams = [beam4,beam5,beam6]
        # while debugging, set this to True
        self.verbose = False
        self.helper = _tmp_test_plan_writer("test.dcm",spotspecs=self.testbeams)
        #self.helper = _test_plan_creator("test.dcm",spotspecs=[beam0])
    def tearDown(self):
        del self.helper
    def test_dicom(self):
        """
        Self consistency check for test dicom file.
        """
        test_rp = _check_rp_dicom_file("test.dcm",self.verbose)
        self.assertEqual(3,len(test_rp.IonBeamSequence))
        self.assertEqual(4,int(test_rp.IonBeamSequence[0].BeamNumber))
        self.assertEqual(5,int(test_rp.IonBeamSequence[1].BeamNumber))
        self.assertEqual(6,int(test_rp.IonBeamSequence[2].BeamNumber))
        self.assertEqual("3.0",test_rp.IonBeamSequence[1].IonControlPointSequence[0].ScanSpotTuneID)
        self.assertAlmostEqual(120.0,test_rp.IonBeamSequence[2].IonControlPointSequence[1].NominalBeamEnergy)
        self.assertAlmostEqual(0.42,test_rp.IonBeamSequence[0].IonControlPointSequence[0].ScanSpotMetersetWeights)
        self.assertAlmostEqual(0.0,test_rp.IonBeamSequence[1].IonControlPointSequence[0].ScanSpotMetersetWeights[0])
        self.assertAlmostEqual(0.6,test_rp.IonBeamSequence[1].IonControlPointSequence[0].ScanSpotMetersetWeights[3])
        self.assertAlmostEqual(0.4,test_rp.IonBeamSequence[2].IonControlPointSequence[1].ScanSpotMetersetWeights[0])
        self.assertAlmostEqual(45.,test_rp.IonBeamSequence[1].IonControlPointSequence[0].PatientSupportAngle)
        self.assertAlmostEqual(60.,test_rp.IonBeamSequence[2].IonControlPointSequence[0].GantryAngle)
    def test_mswtot(self):
        test_rp = _check_rp_dicom_file("test.dcm",self.verbose)
        mswtot_list = _get_mswtot_list(test_rp,self.verbose)
        self.assertEqual(3,len(mswtot_list))
        self.assertAlmostEqual(0.42,mswtot_list[0])
        self.assertAlmostEqual(1.0,mswtot_list[1])
        self.assertAlmostEqual(2.7,mswtot_list[2])
    def test_beam_numbers(self):
        test_rp = _check_rp_dicom_file("test.dcm",self.verbose)
        numbers = _get_beam_numbers(test_rp,self.verbose)
        self.assertEqual([4,5,6],numbers)
    def test_angles_and_iscoCs(self):
        test_rp = _check_rp_dicom_file("test.dcm",self.verbose)
        gantry_angles, patient_angles, iso_centers = _get_angles_and_isoCs(test_rp,self.verbose)
        nbeams=len(self.testbeams)
        self.assertEqual(len(gantry_angles),nbeams)
        self.assertEqual(len(patient_angles),nbeams)
        self.assertEqual(len(iso_centers),nbeams)
        for g,p,i,b in zip(gantry_angles, patient_angles, iso_centers, self.testbeams):
            self.assertAlmostEqual(g,b["G"])
            self.assertAlmostEqual(p,b["P"])
            self.assertEqual(3,len(i))
            for j in range(3):
                self.assertAlmostEqual(i[j],b["I"][j])
    def test_check_output_filename(self):
        with self.assertRaisesRegex(IOError,"suffix"):
            _check_output_filename("error.dcm",self.verbose)
        with self.assertRaisesRegex(IOError,"suffix"):
            _check_output_filename("also_error",self.verbose)
        with self.assertRaisesRegex(IOError,"at least one charachter"):
            _check_output_filename(".txT",self.verbose)
        self.assertEqual("abc",_check_output_filename("abc.txt",self.verbose))
        self.assertEqual("abc",_check_output_filename("/foo/bar/abc.txt",self.verbose))
        self.assertEqual("ab_c",_check_output_filename("ab.c.txt",self.verbose))
        self.assertEqual("_",_check_output_filename(" .txt",self.verbose))
        self.assertEqual("_Box_6__0__0__25___Rashi",_check_output_filename("_Box 6 (0, 0, 25) ^Rashi.txt",self.verbose))
    def test_the_whole_thing_already(self):
        dicom_rt_pbs_plan_to_gate_conversion("test.dcm","zero_tolerance.txt",allow0=True,verbose=self.verbose)
        dicom_rt_pbs_plan_to_gate_conversion("test.dcm","zero_nontolerance.txt",allow0=False,verbose=self.verbose)

class test_workarounds(unittest.TestCase):
    """
    Check that the workarounds for missing information are indeed working.
    """
    def setUp(self):
        self.testbeams = list()
        # now all beams have the same name and number
        # (I saw this in an actual treatment plan, not for a patient but used in commissioning / beam delivery verification.)
        self.testbeams.append(dict(Nm="foo", Nr=-1, # G=90., P=90., I=-1.,
                                   # let's remove the tune ID
                                   controlpoints=[dict(SSW=[0.42e9],E=100., # T="3.0",
                                                       SSP=[(2.,1)])]))
        self.testbeams.append(dict(Nm="foo", Nr=-1, #G=90., P=90., I=-1,
                                   # half of the spots have zero weight
                                   controlpoints=[dict(SSW=[0.,0.4e9,0.,0.6e9],E=110.,T="3.0",
                                                       SSP=[(2.,1.),(2.,1),(3.,4.),(3.,4.)])]))
        self.testbeams.append(dict(Nm="foo", Nr=-1, #G=90., P=90., I=-1,
                                   # All spots in first control point have zero weight
                                   # This actually happens in actual patient plans as well.
                                   controlpoints=[dict(SSW=[0.0,0.0,0.0],E=120.,T="3.1",
                                                       SSP=[(2.,1),(3.,4.),(6.,5.)]),
                                                  dict(SSW=[0.4e9,0.5e9,0.6e9],E=120.,T="3.1",
                                                       SSP=[(2.,1),(3.,4.),(6.,5.)])]))
        # while debugging, set this to True
        self.verbose = False
        self.helper = _tmp_test_plan_writer("test.dcm",spotspecs=self.testbeams)
        #self.helper = _test_plan_creator("test.dcm",spotspecs=[beam0])
    def tearDown(self):
        del self.helper
    def test_dicom(self):
        """
        Self consistency check for test dicom file.
        """
        test_rp = _check_rp_dicom_file("test.dcm",self.verbose)
        self.assertEqual( 3,len(test_rp.IonBeamSequence))
        self.assertEqual(-1,int(test_rp.IonBeamSequence[0].BeamNumber))
        self.assertEqual(-1,int(test_rp.IonBeamSequence[1].BeamNumber))
        self.assertEqual(-1,int(test_rp.IonBeamSequence[2].BeamNumber))
        self.assertEqual("3.1",test_rp.IonBeamSequence[2].IonControlPointSequence[1].ScanSpotTuneID)
        self.assertEqual("",test_rp.IonBeamSequence[0].IonControlPointSequence[0].IsocenterPosition)
        self.assertEqual("",test_rp.IonBeamSequence[1].IonControlPointSequence[0].IsocenterPosition)
        self.assertEqual("",test_rp.IonBeamSequence[2].IonControlPointSequence[0].IsocenterPosition)
        self.assertAlmostEqual(120.0   ,test_rp.IonBeamSequence[2].IonControlPointSequence[1].NominalBeamEnergy)
        self.assertAlmostEqual(  0.42e9,test_rp.IonBeamSequence[0].IonControlPointSequence[0].ScanSpotMetersetWeights)
        self.assertAlmostEqual(  0.0   ,test_rp.IonBeamSequence[1].IonControlPointSequence[0].ScanSpotMetersetWeights[0])
        self.assertAlmostEqual(  0.6e9 ,test_rp.IonBeamSequence[1].IonControlPointSequence[0].ScanSpotMetersetWeights[3])
        self.assertAlmostEqual(  0.5e9 ,test_rp.IonBeamSequence[2].IonControlPointSequence[1].ScanSpotMetersetWeights[1])
        self.assertFalse( hasattr(test_rp.IonBeamSequence[1].IonControlPointSequence[0], "PatientSupportAngle"))
        self.assertFalse( hasattr(test_rp.IonBeamSequence[2].IonControlPointSequence[0], "GantryAngle"))
    def test_mswtot(self):
        test_rp = _check_rp_dicom_file("test.dcm",self.verbose)
        mswtot_list = _get_mswtot_list(test_rp,self.verbose)
        self.assertEqual(3,len(mswtot_list))
        self.assertAlmostEqual(0.42e9,mswtot_list[0])
        self.assertAlmostEqual(1.0e9,mswtot_list[1])
        self.assertAlmostEqual(1.5e9,mswtot_list[2])
    def test_beam_numbers(self):
        test_rp = _check_rp_dicom_file("test.dcm",self.verbose)
        numbers = _get_beam_numbers(test_rp,self.verbose)
        # These numbers are now fixed, right?
        self.assertEqual([1,2,3],numbers)
    def test_angles_and_iscoCs(self):
        test_rp = _check_rp_dicom_file("test.dcm",self.verbose)
        # The "unittest" module has an "assertWarning" test, maybe we should apply that here.
        logger.info("After this message you should see 3x3 warnings about 'absent/corrupted isocenter' and missing patient & gantry angles.")
        gantry_angles, patient_angles, iso_centers = _get_angles_and_isoCs(test_rp,self.verbose)
        for g,p,i in zip(gantry_angles, patient_angles, iso_centers):
            # missing in input, therefore all zero!
            self.assertEqual(g,0.)
            self.assertEqual(p,0.)
            self.assertEqual(3,len(i))
            for j in range(3):
                self.assertEqual(i[j],0.)
    def test_the_whole_thing_already(self):
        # "unittest" has an "assertWarning" test, maybe we should apply that here.
        logger.info("After this message you should see 3x3 warnings about 'absent/corrupted isocenter' and missing patient & gantry angles.")
        dicom_rt_pbs_plan_to_gate_conversion("test.dcm","zero_tolerance.txt",allow0=True,verbose=self.verbose)
        # "unittest" has an "assertWarning" test, maybe we should apply that here.
        logger.info("After this message you should see 3x3 warnings about 'absent/corrupted isocenter' and missing patient & gantry angles.")
        dicom_rt_pbs_plan_to_gate_conversion("test.dcm","zero_nontolerance.txt",allow0=False,verbose=self.verbose)

# vim: set et ts=4 ai sw=4:



# -----------------------------------------------------------------------------
#   Copyright (C): OpenGATE Collaboration
#   This software is distributed under the terms
#   of the GNU Lesser General  Public Licence (LGPL)
#   See LICENSE.md for further details
# -----------------------------------------------------------------------------

#Compare two 3D images using the gamma index formalism as introduced by Daniel Low (1998).

import numpy as np
import itk
import logging
from tqdm import tqdm
logger=logging.getLogger(__name__)

def _reldiff2(dref,dtarget,ddref):
    """
    Convenience function for implementation of the following functions.
    The arguments `dref` and `dtarget` maybe scalars or arrays.
    The calling code is responsible for avoiding division by zero (make sure that ddref>0).
    """
    ddiff=dtarget-dref
    reldd2=(ddiff/ddref)**2
    return reldd2

def get_gamma_index(ref,target,**kwargs):
    """
    Compare two 3D images using the gamma index formalism as introduced by Daniel Low (1998).
    The positional arguments 'ref' and 'target' should behave like ITK image objects.
    Possible keyword arguments include:
    * dd indicates "dose difference" scale as a relative value, in units of percent
      (the dd value is this percentage of the max dose in the reference image)
    * ddpercent is a flag, True (default) means that dd is given in percent, False means that dd is absolute.
    * dta indicates distance scale ("distance to agreement") in millimeter (e.g. 3mm)
    * threshold indicates minimum dose value (exclusive) for calculating gamma values
    * verbose is a flag, True will result in a progress bar. All other chatter goes to the "debug" level.
    Returns an image with the same geometry as the target image.
    For all target voxels in the overlap between ref and target that have d>dmin, a gamma index value is given.
    For all other voxels the "defvalue" is given.
    TODO: allow 2D images, by creating 3D images with a 1-bin Z dimension. Should be very easy.
    The 3D gamma image computed using these "fake 3D" images can then be collapsed back to a 2D image.
    """
    if (np.allclose(ref.GetOrigin(),target.GetOrigin())) and \
       (np.allclose(ref.GetSpacing(),target.GetSpacing())) and \
       (ref.GetLargestPossibleRegion().GetSize() == ref.GetLargestPossibleRegion().GetSize() ):
        logger.debug("Images with equal geometry, using the slightly faster implementation.")
        return gamma_index_3d_equal_geometry(ref,target,**kwargs)
    else:
        logger.debug("Images with different geometry, using the slightly slower implementation.")
        return gamma_index_3d_unequal_geometry(ref,target,**kwargs)


# FIXME: Should this function remain public or be made private (by prefixing it with an _underscore)?
# TODO: Discuss whether to keep this function. It is 30% faster than the
# "unequal geometry" implementation on the same input images. Is that worth it?
def gamma_index_3d_equal_geometry(imgref,imgtarget,dta=3.,dd=3., ddpercent=True,threshold=0.,defvalue=-1.,verbose=False):
    """
    Compare two images with equal geometry, using the gamma index formalism as introduced by Daniel Low (1998).
    * ddpercent indicates "dose difference" scale as a relative value, in units percent (the dd value is this percentage of the max dose in the reference image)
    * ddabs indicates "dose difference" scale as an absolute value
    * dta indicates distance scale ("distance to agreement") in millimeter (e.g. 3mm)
    * threshold indicates minimum dose value (exclusive) for calculating gamma values: target voxels with dose<=threshold are skipped and get assigned gamma=defvalue.
    Returns an image with the same geometry as the target image.
    For all target voxels that have d>threshold, a gamma index value is given.
    For all other voxels the "defvalue" is given.
    If geometries of the input images are not equal, then a `ValueError` is raised.
    """
    aref=itk.array_view_from_image(imgref).swapaxes(0,2)
    atarget=itk.array_view_from_image(imgtarget).swapaxes(0,2)
    if aref.shape != atarget.shape:
        raise ValueError("input images have different geometries ({} vs {} voxels)".format(aref.shape,atarget.shape))
    if not np.allclose(imgref.GetSpacing(),imgtarget.GetSpacing()):
        raise ValueError("input images have different geometries ({} vs {} spacing)".format(imgref.GetSpacing(),imgtarget.GetSpacing()))
    if not np.allclose(imgref.GetOrigin(),imgtarget.GetOrigin()):
        raise ValueError("input images have different geometries ({} vs {} origin)".format(imgref.GetOrigin(),imgtarget.GetOrigin()))
    if ddpercent:
        dd *= 0.01*np.max(aref)
    relspacing = np.array(imgref.GetSpacing(),dtype=float)/dta
    inv_spacing = np.ones(3,dtype=float)/relspacing
    g00=np.ones(aref.shape,dtype=float)*-1
    mask=atarget>threshold
    g00[mask]=np.sqrt(_reldiff2(aref[mask],atarget[mask],dd))
    nx,ny,nz = atarget.shape
    ntot = nx*ny*nz
    nmask = np.sum(mask)
    logger.debug("Both images have {} x {} x {} = {} voxels.".format(nx,ny,nz,ntot))
    logger.debug("{} target voxels have a dose > {}.".format(nmask,threshold))
    g2 = np.zeros((nx,ny,nz),dtype=float)
    if verbose:
        pbar = tqdm(total=nmask, leave=False)
    for x in range(nx):
        for y in range(ny):
            for z in range(nz):
                if g00[x,y,z] < 0:
                    continue
                igmax=np.round(g00[x,y,z]*inv_spacing).astype(int) # maybe we should use "floor" instead of "round"
                if (igmax==0).all():
                    g2[x,y,z]=g00[x,y,z]**2
                else:
                    ixmin = max(x-igmax[0],0)
                    ixmax = min(x+igmax[0]+1,nx)
                    iymin = max(y-igmax[1],0)
                    iymax = min(y+igmax[1]+1,ny)
                    izmin = max(z-igmax[2],0)
                    izmax = min(z+igmax[2]+1,nz)
                    ix,iy,iz = np.meshgrid(np.arange(ixmin,ixmax),
                                           np.arange(iymin,iymax),
                                           np.arange(izmin,izmax),indexing='ij')
                    g2mesh = _reldiff2(aref[ix,iy,iz],atarget[x,y,z],dd)
                    g2mesh += ((relspacing[0]*(ix-x)))**2
                    g2mesh += ((relspacing[1]*(iy-y)))**2
                    g2mesh += ((relspacing[2]*(iz-z)))**2
                    g2[x,y,z] = np.min(g2mesh)
                if verbose:
                    pbar.update(1)
    if verbose:
        pbar.close()
    g=np.sqrt(g2)
    g[np.logical_not(mask)]=defvalue
    # ITK does not support double precision images by default => cast down to float32.
    # Also: only the first few digits of gamma index values are interesting.
    gimg=itk.image_from_array(g.swapaxes(0,2).astype(np.float32).copy())
    gimg.CopyInformation(imgtarget)
    logger.debug(f"Computed {nmask} gamma values assuming EQUAL geometry in target and reference")
    return gimg

# FIXME: should this function remain public or be made private (by prefixing it with an _underscore)?
def gamma_index_3d_unequal_geometry(imgref,imgtarget,dta=3.,dd=3.,ddpercent=True,threshold=0.,defvalue=-1.,verbose=False):
    """
    Compare 3-dimensional arrays with possibly different spacing and different origin, using the
    gamma index formalism, popular in medical physics.
    We assume that the meshes are *NOT* rotated w.r.t. each other.
    * `dd` indicates by default the "dose difference" scale as a relative value,
      in units percent (the dd value is this percentage of the max dose in the reference image).
    * If `ddpercent` is False, then dd is taken as an absolute value.
    * `dta` indicates distance scale ("distance to agreement") in millimeter (e.g. 3mm)
    * `threshold` indicates minimum dose value (exclusive) for calculating gamma values
    Returns an image with the same geometry as the target image.
    For all target voxels that are in the overlap region with the refernce image and that have d>threshold,
    a gamma index value is given. For all other voxels the "defvalue" is given.
    """
    # get arrays
    aref = itk.array_view_from_image(imgref).swapaxes(0,2)
    atarget = itk.array_view_from_image(imgtarget).swapaxes(0,2)
    if ddpercent:
        dd *= 0.01*np.max(aref)
    # test consistency: both must be 3D
    # it would be cool to make this for 2D as well (and D>3), but not now
    if len(aref.shape) != 3 or len(atarget.shape) != 3:
        return None
    #bbref  = bounding_box(imgref)
    #bbtarget = bounding_box(imgtarget)
    areforigin = np.array(imgref.GetOrigin())
    arefspacing = np.array(imgref.GetSpacing())
    atargetorigin = np.array(imgtarget.GetOrigin())
    atargetspacing = np.array(imgtarget.GetSpacing())
    nx,ny,nz = atarget.shape
    mx,my,mz = aref.shape
    ntot = nx*ny*nz
    mtot = mx*my*mz
    dta2  = dta**2
    mask  = atarget>threshold
    nmask=np.sum(mask)
    if nmask==0:
        logger.error("target has no dose over threshold.")
        dummy = itk.image_from_array((np.ones(atarget.shape)*defvalue).swapaxes(0,2).copy())
        dummy.CopyInformation(imgtarget)
        return dummy
    # now define the indices of the target image voxel centers
    ixtarget, iytarget, iztarget = np.meshgrid(np.arange(nx),np.arange(ny),np.arange(nz),indexing='ij')
    xtarget = atargetorigin[0]+ixtarget*atargetspacing[0]
    ytarget = atargetorigin[1]+iytarget*atargetspacing[1]
    ztarget = atargetorigin[2]+iztarget*atargetspacing[2]
    # indices ref image voxel centers that are closest to the target image voxel centers
    ixref = np.round((xtarget-areforigin[0])/arefspacing[0]).astype(int)
    iyref = np.round((ytarget-areforigin[1])/arefspacing[1]).astype(int)
    izref = np.round((ztarget-areforigin[2])/arefspacing[2]).astype(int)
    # keep within range
    overlap = (ixref>=0)*(iyref>=0)*(izref>=0)*(ixref<mx)*(iyref<my)*(izref<mz)
    mask *= overlap
    noverlap = np.sum(overlap)
    nmask=np.sum(mask)
    if nmask==0:
        logger.error("images do not seem to overlap.")
        dummy = itk.image_from_array((np.ones(atarget.shape)*defvalue).swapaxes(0,2).copy())
        dummy.CopyInformation(imgtarget)
        return dummy
    logger.debug("Reference image has {} x {} x {} = {} voxels.".format(mx,my,mz,mtot))
    logger.debug("Target image has {} x {} x {} = {} voxels.".format(nx,ny,nz,ntot))
    logger.debug("{} target voxels are in the intersection of target and reference image.".format(noverlap))
    logger.debug("{} of these have dose > {}.".format(nmask,threshold))
    if verbose:
        pbar = tqdm(total=nmask, leave=False)
    # grid of "close points" in reference image
    xref = areforigin[0]+ixref*arefspacing[0]
    yref = areforigin[1]+iyref*arefspacing[1]
    zref = areforigin[2]+izref*arefspacing[2]
    # get a gamma value on this closest point
    gclose2 = np.zeros([nx,ny,nz],dtype=float)
    gclose2[mask] = _reldiff2(aref[ixref[mask],iyref[mask],izref[mask]],atarget[mask],dd) + \
                                              ((xtarget[mask]-xref[mask])**2 + \
                                               (ytarget[mask]-yref[mask])**2 + \
                                               (ztarget[mask]-zref[mask])**2)/dta2
    gclose = np.array(np.sqrt(gclose2))
    #igclose = np.array(np.ceil(np.sqrt(gclose2)),dtype=int)
    g2=np.zeros([nx,ny,nz],dtype=float)
    #print("going to loop over {} voxels with large enough dose in reference image".format(np.sum(mask)))
    for mixref,miyref,mizref,mixtarget,miytarget,miztarget,mgclose in zip(ixref[mask], iyref[mask], izref[mask],
                                                                    ixtarget[mask],iytarget[mask],iztarget[mask],gclose[mask]):
        #dtarget = atarget[mixtarget,miytarget,miztarget]
        #dref = aref[mixref,miyref,mizref]
        ixyztarget = np.array((mixtarget,miytarget,miztarget))
        ixyzref = np.array((mixref,miyref,mizref))
        targetpos = atargetorigin + ixyztarget*atargetspacing
        refpos = areforigin  + ixyzref*arefspacing
        dixyz = np.floor(mgclose*dta/arefspacing).astype(int) # or round, or ceil?
        imax = np.minimum(ixyzref+dixyz+1,(mx,my,mz))
        imin = np.maximum(ixyzref-dixyz  ,( 0, 0, 0))
        mixnear,miynear,miznear = np.meshgrid(np.arange(imin[0],imax[0]),
                                              np.arange(imin[1],imax[1]),
                                              np.arange(imin[2],imax[2]),
                                              indexing='ij')
        g2near  = _reldiff2(aref[mixnear,miynear,miznear],atarget[mixtarget,miytarget,miztarget],dd)
        g2near += (areforigin[0]+mixnear*arefspacing[0]-targetpos[0])**2/dta2
        g2near += (areforigin[1]+miynear*arefspacing[1]-targetpos[1])**2/dta2
        g2near += (areforigin[2]+miznear*arefspacing[2]-targetpos[2])**2/dta2
        g2[mixtarget,miytarget,miztarget] = np.min(g2near)
        if verbose:
            pbar.update(1)
    g=np.sqrt(g2)
    g[np.logical_not(mask)]=defvalue
    # ITK does not support double precision images by default => cast down to float32.
    # Also: only the first few digits of gamma index values are interesting.
    gimg=itk.image_from_array(g.swapaxes(0,2).astype(np.float32).copy())
    gimg.CopyInformation(imgtarget)
    logger.debug(f"Computed {nmask} gamma values assuming UNEQUAL geometry in target and reference")
    return gimg

#####################################################################################
# TODO: include the unit test in implementation (like here), or have it in a separate test directory?
#####################################################################################
import unittest
import os,sys
from datetime import datetime
from .logging_conf import LoggedTestCase

class Test_GammaIndex3dIdenticalMesh(LoggedTestCase):
    def test_identity(self):
        # two identical images should give gamma=0.0 in all voxels
        logger.debug("test identity")
        logger.debug('Test_GammaIndex3dIdenticalMesh test_identity')
        np.random.seed(1234567)
        a_rnd = np.random.uniform(0.,10.,(4,5,6))
        img1 = itk.image_from_array(a_rnd)
        img2 = itk.image_from_array(a_rnd)
        img_gamma = gamma_index_3d_equal_geometry(img1,img2,dd=3.,dta=2.0)
        self.assertTrue( (itk.array_view_from_image(img_gamma) == 0.).all())
        logger.debug("DONE test identity")
    def test_scaling(self):
        logger.debug("test scaling small")
        # two images identical up to a scaling factor 1.03 should give gamma(3%)<=1.0 in all voxels
        logger.debug('Test_GammaIndex3dIdenticalMesh test_scaling')
        np.random.seed(1234567)
        a_rnd = np.random.uniform(0.,10.,(4,5,6))
        img1 = itk.image_from_array(a_rnd)
        img2 = itk.image_from_array(1.03*a_rnd)
        img_gamma = gamma_index_3d_equal_geometry(img1,img2,dd=3.,dta=2.0)
        self.assertTrue( (itk.array_view_from_image(img_gamma) < 1.0001).all())
        logger.debug("DONE test scaling")
    def test_checkerboards(self):
        logger.debug("test 3D checkerboards: have D=0.25 in even voxels and D=0.75 in odd voxels for ref image, vice versa for test image.")
        #for every voxel in the target image, the neighboring voxels in the ref image has the same dose
        #therefore the gamma index should be equal to spacing/dta for all voxels.
        #logger.debug('Test_GammaIndex3dIdenticalMesh test_checkerboards')
        nx,ny,nz=4,5,6
        ix,iy,iz = np.meshgrid(np.arange(nx,dtype=int),np.arange(ny,dtype=int),np.arange(nz,dtype=int),indexing='ij')
        a_odd  = 0.5*(((ix+iy+iz) % 2) == 1).astype(float)+0.25
        a_even = 0.5*(((ix+iy+iz) % 2) == 0).astype(float)+0.25
        img_odd = itk.image_from_array(a_odd)
        img_even = itk.image_from_array(a_even)
        img_gamma_even_odd = gamma_index_3d_equal_geometry(img_even,img_odd,dd=10.,dta=2.)
        img_gamma_odd_even = gamma_index_3d_equal_geometry(img_odd,img_even,dd=10.,dta=2.)
        self.assertTrue(np.allclose(itk.array_view_from_image(img_gamma_odd_even),itk.array_view_from_image(img_gamma_even_odd)))
        self.assertTrue(np.allclose(itk.array_view_from_image(img_gamma_odd_even),0.5))
        logger.debug("DONE test checkerboards")
    def test_large_image(self):
        logger.debug('Test_GammaIndex3dIdenticalMesh test_large_image')
        for N in [1,2,5,10,20]:
        #for N in [1,2,5,10,20,50]: # requires more patience
        #for N in [1,2,5,10,20,50,100]: # requires even more patience
            tgen = datetime.now()
            img_ref = itk.image_from_array(np.ones((N,N,N),dtype=float))
            img_target = itk.image_from_array(np.random.normal(1.,0.02,(N,N,N)))
            tbefore = datetime.now()
            logger.debug("{}^3 voxels generating images took {}".format(N,tbefore-tgen))
            img_NNN_gamma = gamma_index_3d_equal_geometry(img_ref,img_target,dd=2.,dta=2.0)
            tafter = datetime.now()
            logger.debug("{}^3 voxels calculating gamma took {}".format(N,tafter-tbefore))

class Test_GammaIndex3dUnequalMesh(LoggedTestCase):
    def test_EqualMesh(self):
        # For equal meshes, the "unequalmesh" implementation should give the
        # same results as "equal mesh" implementation.
        logger.debug('Test_GammaIndex3dUnequalMesh test_EqualMesh')
        np.random.seed(71234567)
        for i in range(5):
            logger.debug("{}. comparing implementations with 'equal' and 'unequal' geometry assumptions".format(i))
            nxyz=np.random.randint(25,35,3)
            oxyz=np.random.uniform(-100.,100.,3)
            sxyz=np.random.uniform(0.5,2.5,3)
            img_ref = itk.image_from_array(np.ones(nxyz,dtype=float).swapaxes(0,2).copy())
            img_ref.SetOrigin(oxyz)
            img_ref.SetSpacing(sxyz)
            img_target = itk.image_from_array(np.random.normal(1.,0.05,nxyz).swapaxes(0,2).copy())
            img_target.SetOrigin(oxyz)
            img_target.SetSpacing(sxyz)
            t0 = datetime.now()
            img_gamma_equal = gamma_index_3d_equal_geometry(img_ref,img_target,dd=3.,dta=2.0)
            t1 = datetime.now()
            logger.debug("{}. equal implementation with {} voxels took {}".format(i,np.prod(nxyz),t1-t0))
            img_gamma_unequal = gamma_index_3d_unequal_geometry(img_ref,img_target,dd=3.,dta=2.0)
            t2 = datetime.now()
            logger.debug("{}. unequal implementation with {} voxels took {}".format(i,np.prod(nxyz),t2-t1))
            aeq = itk.array_view_from_image(img_gamma_equal)
            auneq = itk.array_view_from_image(img_gamma_unequal)
            logger.debug("eq min/median/mean/max={}/{}/{}/{}".format(np.min(aeq),np.median(aeq),np.mean(aeq),np.max(aeq)))
            logger.debug("uneq min/median/mean/max={}/{}/{}/{}".format(np.min(auneq),np.median(auneq),np.mean(auneq),np.max(auneq)))
            logger.debug("{} out of {} are close".format(np.sum(np.isclose(aeq,auneq)),np.prod(aeq.shape)))
            logger.debug("eq first 2x2x2: {}".format(aeq[:2,:2,:2]))
            logger.debug("uneq first 2x2x2: {}".format(auneq[:2,:2,:2]))
            logger.debug("eq last 2x2x2: {}".format(aeq[-2:,-2:,-2:]))
            logger.debug("uneq last 2x2x2: {}".format(auneq[-2:,-2:,-2:]))
            self.assertTrue( np.allclose(aeq,auneq) )
            logger.debug("Yay!")
    def test_Shift(self):
        # two images identical up to a translation less than half the spacing should yield a gamma index 
        # equal to the ratio of the length of the translation vector and the DTA.
        logger.debug('Test_GammaIndex3dUnequalMesh test_Shift')
        np.random.seed(1234568)
        for i in range(5):
            logger.debug("shift image no. {}".format(i))
            nxyz=np.random.randint(5,15,3)
            oxyz=np.random.uniform(-100.,100.,3)
            sxyz=np.random.uniform(0.5,2.5,3)
            txyz=np.random.uniform(-0.5,0.5,3)*sxyz
            data = np.random.normal(1.,0.1,nxyz)
            img_ref = itk.image_from_array(data.swapaxes(0,2).copy())
            img_ref.SetSpacing(sxyz)
            img_ref.SetOrigin(oxyz)
            img_target = itk.image_from_array(data.swapaxes(0,2).copy()) # same as ref
            img_target.SetSpacing(sxyz) # same as ref
            img_target.SetOrigin(oxyz+txyz) # translated!
            ddp = 3.0 # %
            dta = 2.0 # mm
            img_gamma = gamma_index_3d_unequal_geometry(img_ref,img_target,dd=ddp,dta=dta)
            agamma = itk.array_view_from_image(img_gamma).swapaxes(0,2)
            gval_expected = np.sqrt( np.sum( (txyz/dta)**2 ) )
            self.assertTrue( np.allclose(agamma,gval_expected) )
            logger.debug("ok #voxels={} gval_exp={}".format(np.prod(nxyz),gval_expected) )
    def test_Gradient(self):
        # Let ref and target be two 3D images that are effectively 1D images (only vary in X)
        # and ref has a much smaller spacing, but its volume includes the target volume.
        # Now it's easy to compute 'by hand' the min(gamma(i,j)) for each target voxel.
        # This tests that the gamma calculation is done correctly wehn it needs
        # to "travel" more than one voxel to get to the "minimum".
        logger.debug('Test_GammaIndex3dUnequalMesh test_Gradient')
        refN=(100,10,10)
        refO=(-50.,0.,0.)
        refS=(1.,1.,1.)
        ixref,iyref,izref = np.meshgrid(np.arange(refN[0]),
                                        np.arange(refN[1]),
                                        np.arange(refN[2]),
                                        indexing='ij')
        logger.debug("ixref shape is {}".format(ixref.shape))
        targetN=(10,3,3)
        ixtarget,iytarget,iztarget = np.meshgrid(np.arange(targetN[0]),
                                                 np.arange(targetN[1]),
                                                 np.arange(targetN[2]),
                                                 indexing='ij')
        np.random.seed(1234569)
        for i in range(5):
            logger.debug("{}th gradient test".format(i))
            ###################
            # generate random REF data
            ###################
            refGRAD = np.random.uniform(-1.,1.)
            refDATA = ixref*refGRAD
            refDATA += 0.5 - np.min(refDATA) # ensure that all dose values are positive
            refOFFSET = refDATA[0,0,0]
            img_ref = itk.image_from_array(refDATA.swapaxes(0,2).copy())
            img_ref.SetSpacing(refS)
            img_ref.SetOrigin(refO)
            ######################
            # generate random TARGET data
            ######################
            # random offset in x (but less than reference spacing)
            targetO=np.random.uniform(-0.5,0.5,3)
            # random x spacing, larger than reference but make sure to stay within ref volume
            sx=np.random.uniform(1.5,2.0)
            targetS=(sx,3.,3.)
            targetGRAD = np.random.uniform(-10.,10.)
            targetDATA = ixtarget*targetGRAD
            targetDATA += 0.5 - np.min(targetDATA) # ensure that all dose values are positive
            targetOFFSET = targetDATA[0,0,0]
            img_target = itk.image_from_array(targetDATA.swapaxes(0,2).copy())
            img_target.SetSpacing(targetS)
            img_target.SetOrigin(targetO)
            #####################################
            # loop over range of gamma parameters
            #####################################
            for ddp in np.arange(1.,5.001):
                for dta in np.arange(1.,5.001):
                    ##########################
                    # calculate expected gamma
                    ##########################
                    ix,iy = np.meshgrid(np.arange(targetN[0]),np.arange(refN[0]),indexing='ij')
                    dr2  = (targetO[0]+ix*targetS[0] - refO[0] - iy*refS[0])**2
                    dr2 += np.sum(targetO[1:]**2)
                    dr2 /= dta**2
                    ddref2 = (np.max(refDATA) * ddp * 0.01)**2
                    dd2 = (targetOFFSET+targetGRAD*ix - refGRAD*iy-refOFFSET)**2 / ddref2
                    # find minimum and take sqrt
                    gamma_1d = np.sqrt(np.min(dd2+dr2,axis=1))
                    # copy to 3d
                    gamma_all = np.zeros(targetN,dtype=float)
                    gamma_all[ixtarget,iytarget,iztarget] = gamma_1d[ixtarget]
                    #################################################
                    # calculate gamma gamma_index_3d_unequal_geometry
                    #################################################
                    img_gamma = gamma_index_3d_unequal_geometry(img_ref,img_target,dd=ddp,dta=dta)
                    logger.debug(type(img_gamma))
                    agamma = itk.array_view_from_image(img_gamma).swapaxes(0,2)
                    self.assertTrue( np.allclose(agamma,gamma_all) )
                    logger.debug("ok ddp={} dta={} refGRAD={} targetGRAD={}".format(ddp,dta,refGRAD,targetGRAD))
            logger.debug("{}th gradient test finished".format(i))

# vim: set et ts=4 ai sw=4:



# -----------------------------------------------------------------------------
#   Copyright (C): OpenGATE Collaboration
#   This software is distributed under the terms
#   of the GNU Lesser General  Public Licence (LGPL)
#   See LICENSE.md for further details
# -----------------------------------------------------------------------------

"""
This module provides basic affine transformation and resampling methods for images.
"""

import os
import itk
import numpy as np
import math
import logging
logger=logging.getLogger(__name__)

def applyTransformation(input=None, like=None, spacinglike=None, matrix=None, newsize=None, neworigin=None, newspacing=None, newdirection=None, force_resample=None, keep_original_canvas=None, adaptive=None, rotation=None, rotation_center=None, translation=None, pad=None, interpolation_mode=None, bspline_order=2):
    
    if like is not None and spacinglike is not None:
        logger.error("Choose between like and spacinglike options")
        sys.exit(1)
    if newspacing is not None and spacinglike is not None:
        logger.error("Choose between newspacing and spacinglike options")
        sys.exit(1)

    if force_resample is None:
        force_resample = False
    if keep_original_canvas is None:
        keep_original_canvas = False
    if force_resample and keep_original_canvas:
        logger.error("Choose between force_resample and keep_original_canvas options")
        sys.exit(1)
    if adaptive is None:
        adaptive = False
    if adaptive and not force_resample:
        logger.error("Be sure to activate force_resample flag with adaptive flag")
        sys.exit(1)

    if force_resample and adaptive and (newspacing is not None or spacinglike is not None) and newsize is not None:
        logger.error("With adaptive flag, choose between spacing and size options")
        sys.exit(1)
    imageDimension = input.GetImageDimension()
    if newsize is None:
        newsize = input.GetLargestPossibleRegion().GetSize()
    if len(newsize) != imageDimension:
        logger.error("Size of newsize is not correct (" + str(imageDimension) + "): " + str(newsize))
        sys.exit(1)
    if newspacing is None:
        newspacing = input.GetSpacing()
    if len(newspacing) != imageDimension:
        logger.error("Size of newspacing is not correct (" + str(imageDimension) + "): " + str(newspacing))
        sys.exit(1)
    if newdirection is None:
        newdirection = input.GetDirection()
    if newdirection.GetVnlMatrix().columns() != imageDimension or newdirection.GetVnlMatrix().rows() != imageDimension:
        logger.error("Size of newdirection is not correct (" + str(imageDimension) + "): " + str(newdirection))
        sys.exit(1)

    if like is not None:
        if like.GetImageDimension() != imageDimension:
            logger.error("Like image does not have the same dimension than input")
            sys.exit(1)
        newsize = like.GetLargestPossibleRegion().GetSize()
        neworigin = like.GetOrigin()
        newspacing = like.GetSpacing()
        newdirection = like.GetDirection()
    if spacinglike is not None:
        if spacinglike.GetImageDimension() != imageDimension:
            logger.error("Spacinglike image does not have the same dimension than input")
            sys.exit(1)
        newspacing = spacinglike.GetSpacing()

    if pad is None:
        pad = 0.0
    if interpolation_mode is None:
        interpolation_mode : "linear"

    if not force_resample and not keep_original_canvas:
        if neworigin is None:
            neworigin = input.GetOrigin()
        changeInfoFilter = itk.ChangeInformationImageFilter.New(Input=input)
        changeInfoFilter.SetOutputSpacing(newspacing)
        changeInfoFilter.SetOutputOrigin(neworigin)
        changeInfoFilter.SetOutputDirection(newdirection)
        changeInfoFilter.ChangeSpacingOn()
        changeInfoFilter.ChangeOriginOn()
        changeInfoFilter.ChangeDirectionOn()
        changeInfoFilter.Update()
        return changeInfoFilter.GetOutput()

    centerImageIndex = itk.ContinuousIndex[itk.D, imageDimension]()
    for i in range(imageDimension):
        centerImageIndex[i] = (input.GetLargestPossibleRegion().GetSize()[i]-1)/2.0
    centerImagePoint = input.TransformContinuousIndexToPhysicalPoint(centerImageIndex)
    centerImageArray = [0]*imageDimension
    for i in range(imageDimension):
        centerImageArray[i] = centerImagePoint[i]
    if rotation_center is None:
        rotation_center = np.zeros(imageDimension)
        for i in range(imageDimension):
            rotation_center[i] = centerImagePoint[i]
    if len(rotation_center) != imageDimension:
        logger.error("Size of rotation_center is not correct (" + str(imageDimension) + "): " + str(rotation_center))
        sys.exit(1)

    rotationMatrix = []
    translationMatrix = []
    if not matrix is None:
        if not rotation is None or not translation is None:
            logger.error("Choose between matrix or rotation/translation, not both")
            sys.exit(1)
        if matrix.GetVnlMatrix().columns() != imageDimension+1 or matrix.GetVnlMatrix().rows() != imageDimension+1:
            logger.error("Size of matrix transformation is not correct (" + str(imageDimension+1) + "): " + str(matrix))
            sys.exit(1)
        if matrix.GetVnlMatrix().columns() == 3 or matrix.GetVnlMatrix().columns() == 4:
            rotationMatrix = itk.matrix_from_array(itk.array_from_matrix(matrix)[:imageDimension, :imageDimension])
        else:
            logger.error("We can transform only 2D and 3D images")
            sys.exit(1)
    else:
        if imageDimension == 2:
            if rotation is None:
                rotation = [0]
            if len(rotation) != 1:
                logger.error("Size of rotation is not correct (1): " + str(rotation))
                sys.exit(1)
        elif imageDimension == 3:
            if rotation is None:
                rotation = [0]*imageDimension
            if len(rotation) != imageDimension:
                logger.error("Size of rotation is not correct (3): " + str(rotation))
                sys.exit(1)
        if translation is None:
            translation = [0]*imageDimension
        if len(translation) != imageDimension:
            logger.error("Size of translation is not correct (" + str(imageDimension) + "): " + str(translation))
            sys.exit(1)
        if imageDimension == 2:
            euler = itk.Euler2DTransform[itk.D].New()
            euler.SetRotation(rotation[0]*math.pi/180.0)
            rotationMatrix = euler.GetMatrix()
        elif imageDimension == 3:
            euler = itk.Euler3DTransform[itk.D].New()
            euler.SetRotation(rotation[0]*math.pi/180.0, rotation[1]*math.pi/180.0, rotation[2]*math.pi/180.0)
            rotationMatrix = euler.GetMatrix()
        else:
            logger.error("We can transform only 2D and 3D images")
            sys.exit(1)

    transform = itk.AffineTransform[itk.D, imageDimension].New()
    transform.SetCenter([0]*imageDimension)
    transform.SetTranslation([0]*imageDimension)
    transform.SetMatrix(rotationMatrix)
    inverseTransform = itk.AffineTransform[itk.D, imageDimension].New()
    transform.GetInverse(inverseTransform)
    if not matrix is None:
        translation = itk.array_from_matrix(matrix)[:imageDimension, imageDimension] - rotation_center + rotationMatrix*rotation_center
    translationMatrix = inverseTransform.GetMatrix()*(centerImageArray - rotation_center) - (centerImageArray - rotation_center) - inverseTransform.GetMatrix()*translation

    inputOrigin = itk.Point[itk.D, imageDimension]()
    for i in range(imageDimension):
        inputOrigin[i] = input.GetOrigin()[i]
    preTranslateFilter = itk.ChangeInformationImageFilter.New(Input=input)
    preTranslateFilter.CenterImageOn()
    preTranslateFilter.Update()

    cornersIndex = [itk.ContinuousIndex[itk.D, imageDimension]() for i in range(2**imageDimension)]
    if imageDimension == 2 or imageDimension == 3:
        cornersIndex[0][0] = -0.5
        cornersIndex[0][1] = -0.5
        if imageDimension == 3:
           cornersIndex[0][2] = -0.5
        cornersIndex[1][0] = input.GetLargestPossibleRegion().GetSize()[0]-0.5
        cornersIndex[1][1] = cornersIndex[0][1]
        if imageDimension == 3:
            cornersIndex[1][2] = cornersIndex[0][2]
        cornersIndex[2][0] = cornersIndex[0][0]
        cornersIndex[2][1] = input.GetLargestPossibleRegion().GetSize()[1]-0.5
        if imageDimension == 3:
            cornersIndex[2][2] = cornersIndex[0][2]
        cornersIndex[3][0] = cornersIndex[1][0]
        cornersIndex[3][1] = cornersIndex[2][1]
        if imageDimension == 3:
            cornersIndex[3][2] = cornersIndex[0][2]
        if imageDimension == 3:
            cornersIndex[4][0] = cornersIndex[0][0]
            cornersIndex[4][1] = cornersIndex[0][1]
            cornersIndex[4][2] = input.GetLargestPossibleRegion().GetSize()[2]-0.5
            cornersIndex[5][0] = cornersIndex[1][0]
            cornersIndex[5][1] = cornersIndex[0][1]
            cornersIndex[5][2] = cornersIndex[4][2]
            cornersIndex[6][0] = cornersIndex[0][0]
            cornersIndex[6][1] = cornersIndex[2][1]
            cornersIndex[6][2] = cornersIndex[4][2]
            cornersIndex[7][0] = cornersIndex[1][0]
            cornersIndex[7][1] = cornersIndex[2][1]
            cornersIndex[7][2] = cornersIndex[4][2]
        outputCorners = np.zeros((2**imageDimension, imageDimension))
        for i in range(2**imageDimension):
            outputCorners[i, :] = inverseTransform.GetMatrix()*preTranslateFilter.GetOutput().TransformContinuousIndexToPhysicalPoint(cornersIndex[i])
        minOutputCorner = np.zeros(imageDimension)
        maxOutputCorner = np.zeros(imageDimension)

        for i in range(imageDimension):
            minOutputCorner[i] = min(outputCorners[:, i])
            maxOutputCorner[i] = max(outputCorners[:, i])
        temp = minOutputCorner + 0.5*itk.array_from_vnl_vector(newspacing.GetVnlVector())
        originAfterRotation = itk.Point[itk.D, imageDimension]()
        for i in range(imageDimension):
            originAfterRotation[i] = temp[i]
        temp = (maxOutputCorner - minOutputCorner)/itk.array_from_vnl_vector(newspacing.GetVnlVector())
        sizeAfterRotation = itk.Size[imageDimension]()
        for i in range(imageDimension):
            sizeAfterRotation[i] = int(math.ceil(temp[i]))
    else:
        logger.error("We can transform only 2D and 3D images")
        sys.exit(1)

    tempImageType = itk.Image[itk.F, imageDimension]
    castImageFilter = itk.CastImageFilter[type(input), tempImageType].New()
    castImageFilter.SetInput(preTranslateFilter.GetOutput())
    castImageFilter.Update()
    
    resampleFilter = itk.ResampleImageFilter.New(Input=castImageFilter.GetOutput())
    resampleFilter.SetOutputSpacing(newspacing)
    resampleFilter.SetOutputOrigin(originAfterRotation)
    resampleDirection = itk.matrix_from_array(np.eye(imageDimension))
    resampleFilter.SetOutputDirection(resampleDirection)
    resampleFilter.SetSize(sizeAfterRotation)
    resampleFilter.SetTransform(transform)
    if interpolation_mode == "NN":
        interpolator = itk.NearestNeighborInterpolateImageFunction[tempImageType, itk.D].New()
    elif interpolation_mode == "BSpline":
        interpolator = itk.BSplineInterpolateImageFunction[tempImageType, itk.D, itk.F].New()
        interpolator.SetSplineOrder(bspline_order)
    else:
        interpolator = itk.LinearInterpolateImageFunction[tempImageType, itk.D].New()
    resampleFilter.SetInterpolator(interpolator)
    resampleFilter.SetDefaultPixelValue(pad)
    resampleFilter.Update()

    postTranslateFilter = itk.ChangeInformationImageFilter.New(Input=resampleFilter.GetOutput())
    postTranslateFilter.SetOutputOrigin(originAfterRotation + centerImagePoint + translationMatrix)
    postTranslateFilter.ChangeOriginOn()
    postTranslateFilter.Update()

    if neworigin is None and not (itk.array_from_matrix(input.GetDirection()) == np.eye(imageDimension)).all():
        neworigin = postTranslateFilter.GetOutput().GetOrigin()
    elif neworigin is None:
        neworigin = inputOrigin
    if len(neworigin) != imageDimension:
        logger.error("Size of neworigin is not correct (" + str(imageDimension) + "): " + str(neworigin))
        sys.exit(1)

    if force_resample and adaptive:
        if (np.array(newspacing) == np.array(input.GetSpacing())).all():
            temp = np.array(sizeAfterRotation)*itk.array_from_vnl_vector(newspacing.GetVnlVector())/np.array(newsize)
            newspacing = itk.Vector[itk.D, imageDimension]()
            for i in range(imageDimension):
                newspacing[i] = temp[i]
        else:
            newsize = itk.Size[imageDimension]()
            for i in range(imageDimension):
                newsize[i] = sizeAfterRotation[i]

    identityTransform = itk.AffineTransform[itk.D, imageDimension].New()
    resampleFilterCanvas = itk.ResampleImageFilter.New(Input=postTranslateFilter.GetOutput())
    resampleFilterCanvas.SetOutputSpacing(newspacing)
    resampleFilterCanvas.SetOutputOrigin(neworigin)
    resampleFilterCanvas.SetOutputDirection(resampleDirection)
    resampleFilterCanvas.SetSize(newsize)
    resampleFilterCanvas.SetTransform(identityTransform)
    if interpolation_mode == "NN":
        interpolator = itk.NearestNeighborInterpolateImageFunction[tempImageType, itk.D].New()
    else:
        interpolator = itk.LinearInterpolateImageFunction[tempImageType, itk.D].New()
    resampleFilterCanvas.SetInterpolator(interpolator)
    resampleFilterCanvas.SetDefaultPixelValue(pad)
    resampleFilterCanvas.Update()
        
    castImageFilter2 = itk.CastImageFilter[tempImageType, type(input)].New()
    castImageFilter2.SetInput(resampleFilterCanvas.GetOutput())
    castImageFilter2.Update()
    
    return castImageFilter2.GetOutput()


    
    

#####################################################################################
import unittest
import sys
from datetime import datetime
import tempfile
import hashlib
import shutil
from .logging_conf import LoggedTestCase

def createImageExample():
    x = np.arange(-10, 10, 1)
    y = np.arange(-12, 15, 1)
    z = np.arange(-13, 10, 1)
    xx, yy, zz = np.meshgrid(x, y, z)
    image = itk.image_from_array(np.int16(xx))
    image.SetOrigin([7, 3.4, -4.6])
    image.SetSpacing([4, 2, 3.6])
    return image

class Test_Affine_Transform(LoggedTestCase):
    def test_change_info(self):
        logger.info('Test_Affine_Transform test_change_info')
        image = createImageExample()
        transformImage = applyTransformation(input=image, neworigin=[-3, 4, -4.6], newspacing=[3, 3, 3])
        tmpdirpath = tempfile.mkdtemp()
        itk.imwrite(transformImage, os.path.join(tmpdirpath, "testAffineTransform.mha"))
        with open(os.path.join(tmpdirpath, "testAffineTransform.mha"),"rb") as fnew:
            bytesNew = fnew.read()
            new_hash = hashlib.sha256(bytesNew).hexdigest()
            self.assertTrue("43cfa7c5c4dadf403fd3555663bf4b2341b69b6797bbae65824abd160a5ef36b" == new_hash)
        shutil.rmtree(tmpdirpath)

    def test_force_resample(self):
        logger.info('Test_Affine_Transform test_force_resample')
        image = createImageExample()
        transformImage = applyTransformation(input=image, force_resample=True, rotation=[43, 12, 278], rotation_center=np.array([12, 56, 23]), translation=[100, -5, 12], pad=-15, interpolation_mode='linear')
        tmpdirpath = tempfile.mkdtemp()
        itk.imwrite(transformImage, os.path.join(tmpdirpath, "testAffineTransform.mha"))
        with open(os.path.join(tmpdirpath, "testAffineTransform.mha"),"rb") as fnew:
            bytesNew = fnew.read()
            new_hash = hashlib.sha256(bytesNew).hexdigest()
            self.assertTrue("98165d7ed3167b393fb65301bfa5a71670142be0421bd609454186c8c4b07333" == new_hash)
        shutil.rmtree(tmpdirpath)

    def test_keep_original_canvas(self):
        logger.info('Test_Affine_Transform test_keep_original_canvas')
        image = createImageExample()
        transformImage = applyTransformation(input=image, keep_original_canvas=True, rotation=[43, 12, 278], rotation_center=np.array([12, 56, 23]), translation=[100, -5, 12], pad=-15, interpolation_mode='NN')
        tmpdirpath = tempfile.mkdtemp()
        itk.imwrite(transformImage, os.path.join(tmpdirpath, "testAffineTransform.mha"))
        with open(os.path.join(tmpdirpath, "testAffineTransform.mha"),"rb") as fnew:
            bytesNew = fnew.read()
            new_hash = hashlib.sha256(bytesNew).hexdigest()
            self.assertTrue("2ee82f72b33f618b23b2dff553385180565318a85b5819bdc48256a3656e64fb" == new_hash)
        shutil.rmtree(tmpdirpath)

    def test_adaptive(self):
        logger.info('Test_Affine_Transform test_adaptive')
        image = createImageExample()
        newspacing = itk.Vector[itk.D, 3]()
        newspacing.Fill(3)
        transformImage = applyTransformation(input=image, force_resample=True, adaptive=True, newspacing=newspacing, pad=-15, interpolation_mode='linear')
        tmpdirpath = tempfile.mkdtemp()
        itk.imwrite(transformImage, os.path.join(tmpdirpath, "testAffineTransform.mha"))
        with open(os.path.join(tmpdirpath, "testAffineTransform.mha"),"rb") as fnew:
            bytesNew = fnew.read()
            new_hash = hashlib.sha256(bytesNew).hexdigest()
            self.assertTrue("1f3446724ce83d9dd5b150e506181e3c92f9b5892b0781383207b1a616da9a17" == new_hash)
        shutil.rmtree(tmpdirpath)



# -----------------------------------------------------------------------------
#   Copyright (C): OpenGATE Collaboration
#   This software is distributed under the terms
#   of the GNU Lesser General  Public Licence (LGPL)
#   See LICENSE.md for further details
# -----------------------------------------------------------------------------

"""
This module provides resizing (pad or crop) for images.
"""

import os
import itk
import numpy as np
import math
import gatetools as gt
import logging
logger=logging.getLogger(__name__)

def imageResize(input=None, newsize=None,  newsize_mm=None, pad=None):
    if input is None:
        logger.error("Set the input")
        sys.exit(1)
    if newsize is None and newsize_mm is None:
        logger.error("Choose between newsize and newsize_mm options")
        sys.exit(1)
    if newsize is not None and newsize_mm is not None:
        logger.error("Choose between newsize and newsize_mm")
        sys.exit(1)

    if pad is None:
        pad = 0.0

    imageDimension = input.GetImageDimension()
    itkSize = itk.Size[imageDimension]()
    if newsize is None:
        for i in range(imageDimension):
            itkSize[i] = int(round(newsize_mm[i]/input.GetSpacing()[i], 0))
    else:
        for i in range(imageDimension):
            itkSize[i] = newsize[i]

    differenceSize = [0]*imageDimension
    for i in range(imageDimension):
        differenceSize[i] = itkSize[i] - input.GetLargestPossibleRegion().GetSize()[i]

    newSize = itk.Size[imageDimension]()
    newOrigin = itk.Point[itk.F, imageDimension]()
    newTranslation = itk.Point[itk.F, imageDimension]()
    for i in range(imageDimension):
        if differenceSize[i] > 0:
            newSize[i] = itkSize[i]
            if differenceSize[i]%2 ==0:
                newTranslation[i] = -differenceSize[i]/2*input.GetSpacing()[i]
            else:
                newTranslation[i] = -(differenceSize[i]+1)/2*input.GetSpacing()[i]
            newOrigin[i] = input.GetOrigin()[i] + newTranslation[i]
        elif differenceSize[i] < 0:
            newSize[i] = itkSize[i]
            if differenceSize[i]%2 ==0:
                newTranslation[i] = -differenceSize[i]/2*input.GetSpacing()[i]
            else:
                newTranslation[i] = -(differenceSize[i]+1)/2*input.GetSpacing()[i]
            newOrigin[i] = input.GetOrigin()[i] + newTranslation[i]
        else:
            newSize[i] = input.GetLargestPossibleRegion().GetSize()[i]
            newTranslation[i] = 0
            newOrigin[i] = input.GetOrigin()[i]
    outputImage = gt.applyTransformation(input, newsize = newSize, force_resample = True, translation = newTranslation, pad = pad)
    outputImage = gt.applyTransformation(outputImage, neworigin = newOrigin)

    return outputImage


#####################################################################################
import unittest
import sys
from datetime import datetime
import tempfile
import hashlib
import shutil
from .logging_conf import LoggedTestCase

def createImageExample():
    x = np.arange(-10, 10, 1)
    y = np.arange(-12, 15, 1)
    z = np.arange(-13, 10, 1)
    xx, yy, zz = np.meshgrid(x, y, z)
    image = itk.image_from_array(np.int16(xx))
    image.SetOrigin([7, 3.4, -4.6])
    image.SetSpacing([4, 2, 3.6])
    return image

class Test_Image_Resize(LoggedTestCase):
    def test_image_resize(self):
        logger.info('Test_Image_resize test_image_resize')
        image = createImageExample()
        itkSize = itk.Size[3]()
        itkSize[0] = 30
        itkSize[1] = 16
        itkSize[2] = 27
        transformImage = imageResize(input=image, newsize=itkSize, pad=30)
        tmpdirpath = tempfile.mkdtemp()
        itk.imwrite(transformImage, os.path.join(tmpdirpath, "testImageResample.mha"))
        with open(os.path.join(tmpdirpath, "testImageResample.mha"),"rb") as fnew:
            bytesNew = fnew.read()
            new_hash = hashlib.sha256(bytesNew).hexdigest()
            self.assertTrue("40c6f1505e19c76a5a12c138cc8028e84be0ae98607aeb0c151a6240600eae48" == new_hash)
        shutil.rmtree(tmpdirpath)
    def test_image_resize_mm(self):
        logger.info('Test_Image_resize test_image_resize_mm')
        image = createImageExample()
        itkSize = itk.Point[itk.F, 3]()
        itkSize[0] = 120
        itkSize[1] = 32
        itkSize[2] = 97.2
        transformImage = imageResize(input=image, newsize_mm=itkSize, pad=30)
        tmpdirpath = tempfile.mkdtemp()
        itk.imwrite(transformImage, os.path.join(tmpdirpath, "testImageResample.mha"))
        itk.imwrite(image, os.path.join(tmpdirpath, "input.mha"))
        with open(os.path.join(tmpdirpath, "testImageResample.mha"),"rb") as fnew:
            bytesNew = fnew.read()
            new_hash = hashlib.sha256(bytesNew).hexdigest()
            self.assertTrue("40c6f1505e19c76a5a12c138cc8028e84be0ae98607aeb0c151a6240600eae48" == new_hash)
        shutil.rmtree(tmpdirpath)



# -----------------------------------------------------------------------------
#   Copyright (C): OpenGATE Collaboration
#   This software is distributed under the terms
#   of the GNU Lesser General  Public Licence (LGPL)
#   See LICENSE.md for further details
# -----------------------------------------------------------------------------

"""
This module provides a tool to have dvh values
"""

import os
import itk
import numpy as np
import math
import logging
import scipy.interpolate
logger=logging.getLogger(__name__)

def createDVH(dose=None, roi=None, bins=1000, label=1, useCm3=False):

    if dose is None:
        logger.error("Set the dose image")
        sys.exit(1)
    if roi is  None:
        logger.error("Set the ROI image")
        sys.exit(1)

    if not np.allclose(dose.GetLargestPossibleRegion().GetSize(), roi.GetLargestPossibleRegion().GetSize()):
        logger.error("images have incompatible size")
        sys.exit(1)
    elif not np.allclose(dose.GetOrigin(), roi.GetOrigin()):
        logger.error("images have incompatible origins")
        sys.exit(1)
    elif not np.allclose(dose.GetSpacing(), roi.GetSpacing()):
        logger.error("images have incompatible spacing")
        sys.exit(1)

    if bins < 1:
        logger.error("bins must be superior or equal to 1")
        sys.exit(1)

    statsRoiFilter = itk.LabelStatisticsImageFilter.New(dose)
    statsRoiFilter.SetLabelInput(roi)
    statsRoiFilter.UseHistogramsOn()
    statsRoiFilter.Update()
    statsRoiFilter.SetHistogramParameters(bins, statsRoiFilter.GetMinimum(label), statsRoiFilter.GetMaximum(label))
    statsRoiFilter.Update()
    histogramRoi = statsRoiFilter.GetHistogram(label)
    volumePercentage = [0]
    doseValues = []
    for i in range(bins):
        volumePercentage.append(volumePercentage[-1] + histogramRoi.GetFrequency(bins-1-i))
        doseValues.append(histogramRoi.GetMeasurement(bins-1-i, 0))
    volumePercentage = np.array(volumePercentage[1:])/histogramRoi.GetTotalFrequency()*100
    volumePercentage = volumePercentage[::-1]
    doseValues = doseValues[::-1]

    if useCm3:
        volumeRoi = statsRoiFilter.GetCount(label)*dose.GetSpacing()[0]*dose.GetSpacing()[1]*dose.GetSpacing()[2]
        volumePercentage = volumePercentage*volumeRoi/100.0

    return (doseValues, volumePercentage)

def computeD(doseValues, volumePercentage, D):
    volumePercentage = np.flip(volumePercentage)
    doseValues = np.flip(doseValues)
    #Remove multiple values:
    indexToRemove = []
    for index in range(1, len(volumePercentage)):
        if volumePercentage[index] == volumePercentage[index-1]:
            indexToRemove += [index-1]
    volumePercentage = np.delete(volumePercentage, indexToRemove)
    doseValues = np.delete(doseValues, indexToRemove)

    doseInterpolated = scipy.interpolate.interp1d(volumePercentage, doseValues, kind='cubic', assume_sorted=False)
    return(doseInterpolated(D))

def computeV(doseValues, volumePercentage, V):
    volumeInterpolated = scipy.interpolate.interp1d(doseValues, volumePercentage, kind='cubic')
    return(volumeInterpolated(V))


#####################################################################################
import unittest
import sys
from datetime import datetime
import tempfile
import hashlib
import shutil
import wget
import pydicom
import gatetools as gt
from matplotlib import pyplot as plt
from .logging_conf import LoggedTestCase

class Test_DVH(LoggedTestCase):
    def test_dvh(self):
        logger.info('Test_DVH test_dvh')
        tmpdirpath = tempfile.mkdtemp()
        filenameStruct = wget.download("https://gitlab.in2p3.fr/opengate/gatetools_data/-/raw/master/rtstruct.dcm?inline=false", out=tmpdirpath, bar=None)
        structset = pydicom.read_file(os.path.join(tmpdirpath, filenameStruct))
        filenameDose = wget.download("https://gitlab.in2p3.fr/opengate/gatetools_data/-/raw/master/rtdose.dcm?inline=false", out=tmpdirpath, bar=None)
        doseImage = gt.read_3d_dicom([os.path.join(tmpdirpath, filenameDose)])
        transformImage = gt.applyTransformation(input=doseImage, neworigin=[-176, -320, -235])

        aroi = gt.region_of_interest(structset, "PTV")
        mask = aroi.get_mask(transformImage, corrected=False)
        doseValues, volumePercentage = createDVH(transformImage, mask, bins=100)
        self.assertTrue(np.allclose(doseValues, np.array([0.10095047205686569, 0.3028514161705971, 0.5047523528337479, 0.7066532969474792, 0.9085542261600494, 1.1104551553726196, 1.3123561143875122, 1.5142570734024048, 1.7161580324172974, 1.9180589318275452, 2.119959831237793, 2.3218607902526855, 2.523761749267578, 2.7256627082824707, 2.9275636672973633, 3.129464626312256, 3.3313655853271484, 3.533266544342041, 3.7351675033569336, 3.9370683431625366, 4.13896918296814, 4.340870141983032, 4.542771100997925, 4.744672060012817, 4.94657301902771, 5.1484739780426025, 5.350374937057495, 5.552275896072388, 5.75417685508728, 5.956077814102173, 6.157978773117065, 6.359879732131958, 6.561780691146851, 6.763681650161743, 6.965582609176636, 7.167483568191528, 7.369384527206421, 7.5712854862213135, 7.773186445236206, 7.9750871658325195, 8.176988124847412, 8.378889083862305, 8.580790042877197, 8.78269100189209, 8.984591960906982, 9.186492919921875, 9.388393878936768, 9.59029483795166, 9.792195796966553, 9.994096755981445, 10.195997714996338, 10.39789867401123, 10.599799633026123, 10.801700592041016, 11.003601551055908, 11.2055025100708, 11.407403469085693, 11.609304428100586, 11.811205387115479, 12.013106346130371, 12.215007305145264, 12.416908264160156, 12.618809223175049, 12.820710182189941, 13.022610664367676, 13.224511623382568, 13.426412582397461, 13.628313541412354, 13.830214500427246, 14.032115459442139, 14.234016418457031, 14.435917377471924, 14.637818336486816, 14.839719295501709, 15.041620254516602, 15.243521213531494, 15.445422172546387, 15.64732313156128, 15.849224090576172, 16.051124572753906, 16.253026008605957, 16.454927444458008, 16.656827926635742, 16.858728408813477, 17.060629844665527, 17.262531280517578, 17.464431762695312, 17.666332244873047, 17.868233680725098, 18.07013511657715, 18.272035598754883, 18.473936080932617, 18.675837516784668, 18.87773895263672, 19.079639434814453, 19.281539916992188, 19.483440399169922, 19.685341835021973, 19.887243270874023, 20.089143753051758])))
        self.assertTrue(np.all(volumePercentage >= 0))
        self.assertTrue(np.all(volumePercentage <= 100))
        self.assertTrue(len(doseValues) == 100)
        self.assertTrue(len(volumePercentage) == 100)
        shutil.rmtree(tmpdirpath)

    def test_dvh_volume(self):
        logger.info('Test_DVH test_dvh_volume')
        tmpdirpath = tempfile.mkdtemp()
        filenameStruct = wget.download("https://gitlab.in2p3.fr/opengate/gatetools_data/-/raw/master/rtstruct.dcm?inline=false", out=tmpdirpath, bar=None)
        structset = pydicom.read_file(os.path.join(tmpdirpath, filenameStruct))
        filenameDose = wget.download("https://gitlab.in2p3.fr/opengate/gatetools_data/-/raw/master/rtdose.dcm?inline=false", out=tmpdirpath, bar=None)
        doseImage = gt.read_3d_dicom([os.path.join(tmpdirpath, filenameDose)])
        transformImage = gt.applyTransformation(input=doseImage, neworigin=[-176, -320, -235])

        aroi = gt.region_of_interest(structset, "PTV")
        mask = aroi.get_mask(transformImage, corrected=False)
        doseValues, volumePercentage = createDVH(transformImage, mask, bins=100, useCm3=True)
        self.assertTrue(np.all(volumePercentage >= 0))
        self.assertTrue(np.all(volumePercentage <= 115120))
        self.assertTrue(len(doseValues) == 100)
        self.assertTrue(len(volumePercentage) == 100)
        shutil.rmtree(tmpdirpath)

    def test_dvh_compute_v(self):
        logger.info('Test_DVH test_dvh_compute_v')
        tmpdirpath = tempfile.mkdtemp()
        filenameStruct = wget.download("https://gitlab.in2p3.fr/opengate/gatetools_data/-/raw/master/rtstruct.dcm?inline=false", out=tmpdirpath, bar=None)
        structset = pydicom.read_file(os.path.join(tmpdirpath, filenameStruct))
        filenameDose = wget.download("https://gitlab.in2p3.fr/opengate/gatetools_data/-/raw/master/rtdose.dcm?inline=false", out=tmpdirpath, bar=None)
        doseImage = gt.read_3d_dicom([os.path.join(tmpdirpath, filenameDose)])
        transformImage = gt.applyTransformation(input=doseImage, neworigin=[-176, -320, -235])

        aroi = gt.region_of_interest(structset, "PTV")
        mask = aroi.get_mask(transformImage, corrected=False)
        doseValues, volumePercentage = createDVH(transformImage, mask)
        self.assertTrue(np.isclose(computeV(doseValues, volumePercentage, 10), 75.33129165068235))
        shutil.rmtree(tmpdirpath)

    def test_dvh_compute_d(self):
        logger.info('Test_DVH test_dvh_compute_d')
        tmpdirpath = tempfile.mkdtemp()
        filenameStruct = wget.download("https://gitlab.in2p3.fr/opengate/gatetools_data/-/raw/master/rtstruct.dcm?inline=false", out=tmpdirpath, bar=None)
        structset = pydicom.read_file(os.path.join(tmpdirpath, filenameStruct))
        filenameDose = wget.download("https://gitlab.in2p3.fr/opengate/gatetools_data/-/raw/master/rtdose.dcm?inline=false", out=tmpdirpath, bar=None)
        doseImage = gt.read_3d_dicom([os.path.join(tmpdirpath, filenameDose)])
        transformImage = gt.applyTransformation(input=doseImage, neworigin=[-176, -320, -235])

        aroi = gt.region_of_interest(structset, "PTV")
        mask = aroi.get_mask(transformImage, corrected=False)
        doseValues, volumePercentage = createDVH(transformImage, mask)
        self.assertTrue(np.isclose(computeD(doseValues, volumePercentage, 95), 0.15428162737918327))
        shutil.rmtree(tmpdirpath)




# -----------------------------------------------------------------------------
#   Copyright (C): OpenGATE Collaboration
#   This software is distributed under the terms
#   of the GNU Lesser General  Public Licence (LGPL)
#   See LICENSE.md for further details
# -----------------------------------------------------------------------------

"""
This module provides basic affine transformation and resampling methods for images.
"""

import gatetools as gt
import itk
import numpy as np
import logging
logger=logging.getLogger(__name__)

def imageStatistics(input=None, mask=None, resample=False, histogramBins=1000):

    if input is None:
        logger.error("Set an input")
        sys.exit(1)

    inputArray = itk.array_from_image(input)
    outputStats = {}
    outputStats["nbPixel"] = inputArray.size
    if not mask is None:
        if resample:
            mask = gt.applyTransformation(input=mask, like=input, force_resample=True)
        if not np.allclose(mask.GetSpacing(), input.GetSpacing()):
            logger.error("Input and mask do not have the same spacing")
            sys.exit(1)
        if not np.allclose(mask.GetOrigin(), input.GetOrigin()):
            logger.error("Input and mask do not have the same origin")
            sys.exit(1)
        if not np.allclose(itk.array_from_matrix(mask.GetDirection()), itk.array_from_matrix(input.GetDirection())):
            logger.error("Input and mask do not have the same direction")
            sys.exit(1)
        if not np.allclose(mask.GetLargestPossibleRegion().GetSize(),  input.GetLargestPossibleRegion().GetSize()):
            logger.error("Input and mask do not have the same size")
            sys.exit(1)

        maskArray = itk.array_from_image(mask)
        if len(np.where(maskArray > 1)[0]) >0:
            logger.error("The mask seems to be a non-binary image")
        index = np.where(maskArray == 1)
        outputStats["nbPixel"] = len(index[0])
        inputArray = inputArray[index]

    outputStats["minimum"] = np.amin(inputArray)
    outputStats["maximum"] = np.amax(inputArray)
    outputStats["sum"] = np.sum(inputArray)
    outputStats["median"] = np.median(inputArray)
    outputStats["mean"] = np.mean(outputStats["sum"]/outputStats["nbPixel"])
    outputStats["variance"] = np.var(inputArray)
    outputStats["sigma"] = np.sqrt(outputStats["variance"])
    outputStats["hist"] = np.histogram(inputArray, histogramBins)

    return outputStats
    

#####################################################################################
import unittest
import sys
from datetime import datetime
import tempfile
import hashlib
import shutil
from .logging_conf import LoggedTestCase

def createImageExample():
    x = np.arange(-10, 10, 1)
    y = np.arange(-12, 15, 1)
    z = np.arange(-13, 10, 1)
    xx, yy, zz = np.meshgrid(x, y, z)
    image = itk.image_from_array(np.int16(xx))
    image.SetOrigin([7, 3.4, -4.6])
    image.SetSpacing([4, 2, 3.6])
    return image

class Test_Image_Statistics(LoggedTestCase):
    def test_image_statistics(self):
        logger.info('Test_Image_Statistics test_image_statistics')
        image = createImageExample()
        outputStats = imageStatistics(input=image, histogramBins=10)
        self.assertTrue(outputStats["nbPixel"] == 12420)
        self.assertTrue(outputStats["mean"] == -0.5)
        self.assertTrue(outputStats["median"] == -0.5)
        self.assertTrue(outputStats["sum"] == -6210)
        self.assertTrue(outputStats["minimum"] == -10)
        self.assertTrue(outputStats["maximum"] == 9)
        self.assertTrue(outputStats["variance"] == 33.25)
        self.assertTrue(outputStats["sigma"] == 5.766281297335398)
        self.assertTrue(np.allclose(outputStats["hist"][0], [1242, 1242, 1242, 1242, 1242, 1242, 1242, 1242, 1242, 1242]))
        self.assertTrue(np.allclose(outputStats["hist"][1], [-10. , -8.1, -6.2, -4.3, -2.4, -0.5, 1.4, 3.3, 5.2, 7.1, 9.]))

    def test_image_statistics_with_mask(self):
        logger.info('Test_Image_Statistics test_image_statistics_with_mask')
        image = createImageExample()
        mask = createImageExample()
        x = np.arange(-10, 10, 1)
        y = np.arange(-12, 15, 1)
        z = np.arange(-13, 10, 1)
        xx, yy, zz = np.meshgrid(x, y, z)
        mask = np.int16(xx)
        mask[:] = 0
        mask[:, 11, :] = 1
        mask = itk.image_from_array(mask)
        mask.SetOrigin([7, 3.4, -4.6])
        mask.SetSpacing([4, 2, 3.6])
        outputStats = imageStatistics(input=image, mask=mask, histogramBins=5)
        self.assertTrue(outputStats["nbPixel"] == 621)
        self.assertTrue(outputStats["mean"] == 1)
        self.assertTrue(outputStats["median"] == 1)
        self.assertTrue(outputStats["sum"] == 621)
        self.assertTrue(outputStats["minimum"] == 1)
        self.assertTrue(outputStats["maximum"] == 1)
        self.assertTrue(outputStats["variance"] == 0)
        self.assertTrue(outputStats["sigma"] == 0)
        self.assertTrue(np.allclose(outputStats["hist"][0], [0, 0, 621, 0, 0]))
        self.assertTrue(np.allclose(outputStats["hist"][1], [0.5, 0.7, 0.9, 1.1, 1.3, 1.5]))

    def test_image_statistics_with_mask_and_resample(self):
        logger.info('Test_Image_Statistics test_image_statistics_with_mask')
        image = createImageExample()
        mask = createImageExample()
        x = np.arange(-10, 10, 1)
        y = np.arange(-12, 15, 1)
        z = np.arange(-13, 10, 1)
        xx, yy, zz = np.meshgrid(x, y, z)
        mask = np.int16(xx)
        mask[:] = 0
        mask[:, 11, :] = 1
        mask = itk.image_from_array(mask)
        mask.SetOrigin([3, 3.4, -4.6])
        mask.SetSpacing([4, 2, 3.6])
        outputStats = imageStatistics(input=image, mask=mask, resample=True, histogramBins=7)
        self.assertTrue(outputStats["nbPixel"] == 594)
        self.assertTrue(outputStats["mean"] == 1)
        self.assertTrue(outputStats["median"] == 1)
        self.assertTrue(outputStats["sum"] == 594)
        self.assertTrue(outputStats["minimum"] == 1)
        self.assertTrue(outputStats["maximum"] == 1)
        self.assertTrue(outputStats["variance"] == 0)
        self.assertTrue(outputStats["sigma"] == 0)
        self.assertTrue(np.allclose(outputStats["hist"][0], [0, 0, 0, 594, 0, 0, 0]))
        self.assertTrue(np.allclose(outputStats["hist"][1], [0.5, 0.64285714, 0.78571429, 0.92857143, 1.07142857, 1.21428571, 1.35714286, 1.5]))



# -----------------------------------------------------------------------------
#   Copyright (C): OpenGATE Collaboration
#   This software is distributed under the terms
#   of the GNU Lesser General  Public Licence (LGPL)
#   See LICENSE.md for further details
# -----------------------------------------------------------------------------

"""
This module contains the default logging configuration for the Gate tools.
It is used by the command line tools and you *can* use it for your own
python scripts based on the gatetools modules. You can of course also define
your own logging configuration.
"""

# -----------------------------------------------------------------------------
#   Copyright (C): OpenGATE Collaboration
#   This software is distributed under the terms
#   of the GNU Lesser General  Public Licence (LGPL)
#   See LICENSE.md for further details
# -----------------------------------------------------------------------------


import logging
from datetime import datetime
import os, sys
import click

_logging_is_already_configured = False

def logging_conf(verbose=False,logfile=""):
    """
    This function is supposed to be called only once, inside the main function
    of your application.

    By default, the logging level is high (i.e. low verbosity) and no logging file
    will be produced.

    You can choose to increase the terminal output (verbose=True). This lowers
    the log level to "INFO". This is not the lowest possible logging level.
    With debugging mode off, the logging messages will be displayed without any
    prefix, i.e. as if they were printed with the "print()" function. With
    debugging mode on, then you get a lengthy prefix that shows you details
    like the module name, line number, date and time.  Terminal output and
    logging file get the same prefix, in order to facilitate finding back
    events in the log file.

    The debugging mode is enabled when the `logfile` argument is a non-empty
    string containing the desired file path the log file (in a directory where
    the user should have write permissions). If the file already exists, then
    the logging information will be *appended*. If the `logfile` argument is
    equal to "auto" then a log file name in the current working directory will
    be generated, composed of the basename of the main script and a prefix
    ".PID.YYMMDD.HHMMSS.log", where PID, YYMMDD and HHMMSS are the process ID,
    date and time, respectively.

    Note that the `verbose` flag only controls how much information is streamed
    to terminal; the log file always contains the full debug-level information.
    """
    global _logging_is_already_configured
    if _logging_is_already_configured:
        logger=logging.getLogger(__name__)
        logger.debug("attempt to re-configure logging ignored...")
        return

    stdout_loglevel = logging.INFO if verbose else logging.WARNING
    if not bool(logfile):
        logging.basicConfig(format='%(message)s',level=stdout_loglevel)
        return
    else:
        logger = logging.getLogger() # get root logger
        logger.setLevel(logging.DEBUG)
        # same format for terminal and for log file
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(lineno)d - %(levelname)s - %(message)s')
        # terminal output
        sh = logging.StreamHandler()
        sh.setFormatter(formatter)
        sh.setLevel(stdout_loglevel)
        logger.addHandler(sh)
        # log file
        if logfile == "auto":
            logfiledir = os.path.realpath(os.getcwd())
            logfilename = os.path.basename(sys.argv[0])
            logfilename += "."+str(os.getpid())+datetime.now().strftime(".%Y%m%d.%H%M%S.log")
            logfilepath = os.path.join(logfiledir,logfilename)
        else:
            logfilepath = os.path.realpath(logfile)
        # check disk space before trying to open log file ("df" = "disk free")
        # TODO: if it turns out that `psutil` is available on practically all systems
        # then we can maybe use that here too, it has a nicer interface.
        vfs=os.statvfs(os.path.dirname(logfilepath))
        MiB=1024*1024
        GiB=1024*MiB
        df = vfs.f_bsize*vfs.f_bavail
        if df < 100*MiB: # totally arbitrary minimum
            raise RuntimeError("Looks like you are very low on disk space, abort...")
        try:
            fh = logging.FileHandler(logfilepath)
            fh.setLevel(logging.DEBUG)
            fh.setFormatter(formatter)
            logger.addHandler(fh)
        except Exception as e:
            logger.error("failed to open debugging log file {}".format(os.path.realpath(logfilepath)))
            logger.error("got exception: '{}'".format(e))
            logger.exception("stack trace:")
            logger.critical("please make sure that the log file path is correct, that there is space on the device and that you have write permissions on it.")
            raise
        # logging is now configured
        # start with logging some system basics that could possibly be useful for debugging
        logger.debug("start logging")
        logger.debug("OS: {}".format(os.uname()))
        logger.debug("python version: {}".format(sys.version))
        logger.debug("#cpus: {}".format(os.cpu_count()))
        logger.debug("load averages: {}".format(str(os.getloadavg())))
        logger.debug("free disk space in log file directory: {0:.2f} GiB ({1:d} MiB)".format(df/GiB,int(df/MiB)))
        logger.debug("full logfile path: {}".format(logfilepath))
        logger.debug("cwd: {}".format(os.getcwd()))
        logger.debug("script: {}".format(sys.argv[0]))
        try:
            import itk
            logger.debug("ITK: {}".format(itk.Version.GetITKVersion()))
            import pydicom
            logger.debug("pydicom: {}".format(".".join(pydicom.__version_info__)))
            import numpy
            logger.debug("numpy: {}".format(numpy.__version__))
            import psutil
            meminfo=psutil.virtual_memory()
            logger.debug("RAM: total={} MiB, available={} MiB ({} % used)".format(meminfo.total/MiB,meminfo.available/MiB,meminfo.percent))
            # TODO: should we emit a warning if we see that available RAM is very low?
        except ImportError as ie:
            logger.warning("oopsie: {}".format(ie))
            logger.warning("trying to continue, keep fingers crossed...")
    _logging_is_already_configured = True



def add_options(options):
    '''
    Function to add all common options to click.

    Current options are --verbose and --log
    '''
    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func
    return _add_options

common_options= [
    click.option('--verbose/--quiet','-v/-q', default=False,
                 help='Be verbose (-v) or keep quiet (-q, default)'),
    click.option('--logfile', required=False, default="",
                 help='Path of debugging log file. Default: empty (no log file is written). If you specify "auto" then an informative logfile name will be generated automatically.')
]


###############################################################################
import unittest
class LoggedTestCase(unittest.TestCase):
    def __init__(self,*args,**kwargs):
        logfilename=f'unittest.{os.getpid()}.log'
        logging_conf(verbose=False, logfile=logfilename)
        unittest.TestCase.__init__(self,*args,**kwargs)



# -----------------------------------------------------------------------------
#   Copyright (C): OpenGATE Collaboration
#   This software is distributed under the terms
#   of the GNU Lesser General  Public Licence (LGPL)
#   See LICENSE.md for further details
# -----------------------------------------------------------------------------

"""

This module provides a function to crop image

"""

# -----------------------------------------------------------------------------
#   Copyright (C): OpenGATE Collaboration
#   This software is distributed under the terms
#   of the GNU Lesser General  Public Licence (LGPL)
#   See LICENSE.md for further details
# -----------------------------------------------------------------------------


import gatetools as gt
import numpy as np
import tqdm
import logging
logger=logging.getLogger(__name__)

def unicity(root_keys):
    """
    Return an array containing the keys of the root file only one (without the version number)
    """
    root_array = []
    for key in root_keys:
        name = key.decode("utf-8").split(";")[0]
        if not name in root_array:
            root_array.append(name)
    return(root_array)

def merge_root(rootfiles, outputfile, incrementRunId=False):
    """
    Merge root files in output files
    """
    try:
        import uproot3 as uproot
    except:
        print("uproot3 is mandatory to merge root file. Please, do:")
        print("pip install uproot3")

    out = uproot.recreate(outputfile)

    #Previous ID values to be able to increment runIn or EventId
    previousId = {}

    #create the dict reading all input root files
    trees = {}
    pbar = tqdm.tqdm(total = len(rootfiles))
    for file in rootfiles:
        root = uproot.open(file)
        root_keys = unicity(root.keys())
        for tree in root_keys:
            if hasattr(root[tree], 'keys'):
                if not tree in trees:
                    trees[tree] = {}
                    trees[tree]["rootDictType"] = {}
                    trees[tree]["rootDictValue"] = {}
                    previousId[tree] = {}
                for branch in root[tree].keys():
                    array = root[tree].array(branch)
                    if len(array) > 0:
                        if type(array[0]) is type(b'c'):
                            array = np.array([0 for xi in array])
                        if not branch in trees[tree]["rootDictType"]:
                            trees[tree]["rootDictType"][branch] = type(array[0])
                            trees[tree]["rootDictValue"][branch] = np.array([])
                        if (not incrementRunId and branch.decode('utf-8').startswith('eventID')) or (incrementRunId and branch.decode('utf-8').startswith('runID')):
                            if not branch in previousId[tree]:
                                previousId[tree][branch] = 0
                            array += previousId[tree][branch]
                            previousId[tree][branch] = max(array) +1
                        trees[tree]["rootDictValue"][branch] = np.append(trees[tree]["rootDictValue"][branch], array)
        pbar.update(1)
    pbar.close()

    #Set the dict in the output root file
    for tree in trees:
        if not trees[tree]["rootDictValue"] == {} or not trees[tree]["rootDictType"] == {}:
            out[tree] = uproot.newtree(trees[tree]["rootDictType"])
            out[tree].extend(trees[tree]["rootDictValue"])




#####################################################################################
import unittest
import tempfile
import wget
import os
import shutil
from .logging_conf import LoggedTestCase

class Test_MergeRoot(LoggedTestCase):
    def test_merge_root_phsp(self):
        try:
            import uproot3 as uproot
        except:
            print("uproot3 is mandatory to merge root file. Please, do:")
            print("pip install uproot3")

        logger.info('Test_MergeRoot test_merge_root_phsp')
        tmpdirpath = tempfile.mkdtemp()
        filenameRoot = wget.download("https://gitlab.in2p3.fr/opengate/gatetools_data/-/raw/master/phsp.root?inline=false", out=tmpdirpath, bar=None)
        gt.merge_root([filenameRoot, filenameRoot],  os.path.join(tmpdirpath, "output.root"))
        input = uproot.open(filenameRoot)
        output = uproot.open(os.path.join(tmpdirpath, "output.root"))
        self.assertTrue(output.keys() == input.keys())
        inputTree = input[input.keys()[0]]
        outputTree = output[output.keys()[0]]
        self.assertTrue(outputTree.keys() == inputTree.keys())
        inputBranch = inputTree.array(inputTree.keys()[1])
        outputBranch = outputTree.array(outputTree.keys()[1])
        self.assertTrue(2*len(inputBranch) == len(outputBranch))
        shutil.rmtree(tmpdirpath)

    def test_merge_root_pet_incrementEvent(self):
        try:
            import uproot3 as uproot
        except:
            print("uproot3 is mandatory to merge root file. Please, do:")
            print("pip install uproot3")

        logger.info('Test_MergeRoot test_merge_root_pet')
        tmpdirpath = tempfile.mkdtemp()
        filenameRoot = wget.download("https://gitlab.in2p3.fr/opengate/gatetools_data/-/raw/master/pet.root?inline=false", out=tmpdirpath, bar=None)
        gt.merge_root([filenameRoot, filenameRoot],  os.path.join(tmpdirpath, "output.root"))
        input = uproot.open(filenameRoot)
        output = uproot.open(os.path.join(tmpdirpath, "output.root"))
        inputTree = input[input.keys()[0]]
        outputTree = output[output.keys()[0]]
        inputRunBranch = inputTree.array(inputTree.keys()[0])
        outputRunBranch = outputTree.array(outputTree.keys()[0])
        self.assertTrue(max(inputRunBranch) == max(outputRunBranch))
        self.assertTrue(2*len(inputRunBranch) == len(outputRunBranch))
        inputEventBranch = inputTree.array(inputTree.keys()[1])
        outputEventBranch = outputTree.array(outputTree.keys()[1])
        self.assertTrue(2*max(inputEventBranch)+1 == max(outputEventBranch))
        self.assertTrue(2*len(inputEventBranch) == len(outputEventBranch))
        shutil.rmtree(tmpdirpath)

    def test_merge_root_pet_incrementRun(self):
        try:
            import uproot3 as uproot
        except:
            print("uproot3 is mandatory to merge root file. Please, do:")
            print("pip install uproot3")

        logger.info('Test_MergeRoot test_merge_root_pet')
        tmpdirpath = tempfile.mkdtemp()
        print(tmpdirpath)
        filenameRoot = wget.download("https://gitlab.in2p3.fr/opengate/gatetools_data/-/raw/master/pet.root?inline=false", out=tmpdirpath, bar=None)
        gt.merge_root([filenameRoot, filenameRoot],  os.path.join(tmpdirpath, "output.root"), True)
        input = uproot.open(filenameRoot)
        output = uproot.open(os.path.join(tmpdirpath, "output.root"))
        inputTree = input[input.keys()[0]]
        outputTree = output[output.keys()[0]]
        inputRunBranch = inputTree.array(inputTree.keys()[0])
        outputRunBranch = outputTree.array(outputTree.keys()[0])
        self.assertTrue(2*max(inputRunBranch)+1 == max(outputRunBranch))
        self.assertTrue(2*len(inputRunBranch) == len(outputRunBranch))
        inputEventBranch = inputTree.array(inputTree.keys()[1])
        outputEventBranch = outputTree.array(outputTree.keys()[1])
        self.assertTrue(max(inputEventBranch) == max(outputEventBranch))
        self.assertTrue(2*len(inputEventBranch) == len(outputEventBranch))
        #shutil.rmtree(tmpdirpath)




from box import Box


def get_pet_data(root_file):
    data = Box()
    try:
        data_pet = root_file['pet_data']
        data.total_nb_primaries = data_pet['total_nb_primaries'].array(library='numpy')[0]
        data.latest_event_ID = data_pet['latest_event_ID'].array(library='numpy')[0]
        data.stop_time_sec = data_pet['stop_time_sec'].array(library='numpy')[0]
        data.start_time_sec = data_pet['start_time_sec'].array(library='numpy')[0]
        return data
    except:
        return False


def get_pet_counts(root_file):
    data = Box()

    coincidences = root_file['Coincidences']
    data.prompts_count = coincidences.num_entries

    delays = root_file['Delay']
    data.delays_count = delays.num_entries

    # Root tree analysis principle:
    # 1) consider a "Tree" (such as root_file['Coincidences'])
    # 2) convert the branches to numpy array
    # 3) use mask to select only some part of the tree

    # Compute the number of Random Coincidences
    # (when the two singles in coincidence came from two different events or are noise)
    event_id1 = coincidences['eventID1'].array(library='numpy')
    event_id2 = coincidences['eventID2'].array(library='numpy')
    # Warning noise events have eventID == -2
    # https://opengate.readthedocs.io/en/latest/digitizer_and_detector_modeling.html?#noise
    mask_trues = (event_id1 == event_id2) & (event_id1 >= 0)
    data.randoms_count = data.prompts_count - len(event_id1[mask_trues])

    # Scattered coincidences (among the true)
    scatter_compton1 = coincidences['comptonPhantom1'].array(library='numpy')
    scatter_compton2 = coincidences['comptonPhantom2'].array(library='numpy')
    scatter_rayleigh1 = coincidences['RayleighPhantom1'].array(library='numpy')
    scatter_rayleigh2 = coincidences['RayleighPhantom2'].array(library='numpy')
    # if any of the value is greater than zero, it means a compton or a rayleigh occurs,
    # so this is a scattered event
    mask_scatter = (scatter_compton1 > 0) | (scatter_compton2 > 0) | (scatter_rayleigh1 > 0) | (scatter_rayleigh2 > 0)
    data.scatter_count = len(scatter_compton1[mask_trues & mask_scatter])

    # Remaining true events
    data.trues_count = data.prompts_count - data.randoms_count - data.scatter_count

    return data



# -----------------------------------------------------------------------------
#   Copyright (C): OpenGATE Collaboration
#   This software is distributed under the terms
#   of the GNU Lesser General  Public Licence (LGPL)
#   See LICENSE.md for further details
# -----------------------------------------------------------------------------

"""

This module provides a function to convert image from one type to another:

"""

# -----------------------------------------------------------------------------
#   Copyright (C): OpenGATE Collaboration
#   This software is distributed under the terms
#   of the GNU Lesser General  Public Licence (LGPL)
#   See LICENSE.md for further details
# -----------------------------------------------------------------------------


import itk
import pydicom
from pydicom.tag import Tag
import gatetools as gt
import numpy as np
import logging
logger=logging.getLogger(__name__)


class dicom_properties:
    def __init__(self):
        self.spacing = [1.0, 1.0, 1.0]
        self.origin = [0.0, 0.0, 0.0]
        self.io = [1.0, 0.0, 0.0, 0.0, 1.0, 0.0]
        self.rs = 1.0
        self.ri = 0.0
        self.image_shape = []

    def read_dicom_slop_intercept(self, slice):
        if Tag(0x28, 0x1052) in slice:
            self.ri = slice[0x28, 0x1052].value #Rescale Intercept
            self.rs = slice[0x28, 0x1053].value #Rescale Slope
        elif Tag(0x11, 0x103b) in slice:
            self.rs = slice[0x11, 0x103b].value #Pixel Scale
            self.ri = slice[0x11, 0x103c].value #Pixel Offset
        elif Tag(0x40, 0x9096) in slice and Tag(0x40, 0x9224) in slice[0x40, 0x9096][0]:
            self.ri = slice[0x40, 0x9096][0][0x40, 0x9224].value #Rescale Intercept
            self.rs = slice[0x40, 0x9096][0][0x40, 0x9225].value #Rescale Slope
        elif Tag(0x3004, 0x000e) in slice:
            self.rs = slice[0x3004, 0x000e].value #Rescale Slope
        if hasattr(self.ri, '__len__'):
            self.ri = self.ri[0]
        if hasattr(self.rs, '__len__'):
            self.rs = self.rs[0]

    def read_dicom_properties(self, slice, nextSlice=None):
        # pixel aspects, assuming all slices are the same

        ps = [1.0, 1.0]
        if Tag(0x28, 0x30) in slice:
            ps = slice.PixelSpacing
        ss = None
        if Tag(0x18, 0x88) in slice:
            ss = abs(float(slice[(0x0018,0X0088)].value))
        if ss == '' or ss is None:
            if Tag(0x3004, 0x000c) in slice:
                ss = slice[0x3004, 0x000c][1] - slice[0x3004, 0x000c][0]
        if ss == '' or ss is None:
            if not nextSlice is None and Tag(0x0020, 0x0032) in slice and Tag(0x0020, 0x0032) in nextSlice:
                ss = abs(nextSlice[0x0020, 0x0032][2] - slice[0x0020, 0x0032][2])
            if ss == '' or ss is None:
                ss = 1.0
        self.spacing = [ps[1], ps[0], ss]
        ip = [0.0, 0.0, 0.0]
        if Tag(0x20, 0x32) in slice:
            ip = slice[0x20, 0x32].value #Image Position
        elif Tag(0x54, 0x22) in slice and Tag(0x20, 0x32) in slice[0x54, 0x22][0]:
            ip = slice[0x54, 0x22][0][0x20, 0x32].value #Image Position
        self.origin = [ip[0], ip[1], ip[2]]
        if Tag(0x20, 0x37) in slice:
            self.io = slice[0x20, 0x37].value #Image Orientation
        elif Tag(0x54, 0x22) in slice and Tag(0x20, 0x37) in slice[0x54, 0x22][0]:
            self.io = slice[0x54, 0x22][0][0x20, 0x37].value #Image Orientation
        if self.io == "" or self.io == None:
            self.io = [1.0, 0.0, 0.0, 0.0, 1.0, 0.0]
        #orientation = [io[0], io[1], io[2], io[3], io[4], io[5]]
        self.read_dicom_slop_intercept(slice)


        self.img_shape = list(slice.pixel_array.shape)

def separate_series(dicomFiles):
    """
    Read dicom and return a dictionary with the different series separated
    """
    files = {}
    #Load dicom files
    for file in dicomFiles:
        try:
            seriesInstanceUID = pydicom.read_file(file)[0x0020, 0x000e].value
        except pydicom.errors.InvalidDicomError:
            ds = pydicom.read_file(file, force=True)
            ds.file_meta.TransferSyntaxUID = pydicom.uid.ImplicitVRLittleEndian
            seriesInstanceUID = ds[0x0020, 0x000e].value
        if seriesInstanceUID not in files:
            files[seriesInstanceUID] = []
        files[seriesInstanceUID].append(file)
    return files

def read_dicom(dicomFiles):
    """

    Read dicom files and return a float 3D image
    """
    files = []
    #Load dicom files
    for file in dicomFiles:
        try:
            files.append(pydicom.read_file(file))
        except pydicom.errors.InvalidDicomError:
            ds = pydicom.read_file(file, force=True)
            ds.file_meta.TransferSyntaxUID = pydicom.uid.ImplicitVRLittleEndian
            files.append(ds)

    # skip files with no SliceLocation (eg scout views)
    slices = []
    skipcount = 0
    if len(files) > 1:
        for f in files:
            if hasattr(f, 'SliceLocation') or hasattr(f, 'ImagePositionPatient'):
                slices.append(f)
            else:
                skipcount = skipcount + 1

        if skipcount >0:
            logger.info("skipped, no SliceLocation: {}".format(skipcount))

        #Remove images with same SopInstanceUID
        noDuplicateSlices = []
        tmpSopInstanceUID = []
        for slice in slices:
            if not slice[0x0008, 0x0018].value in tmpSopInstanceUID:
                noDuplicateSlices.append(slice)
                tmpSopInstanceUID.append(slice[0x0008, 0x0018].value)

        # ensure they are in the correct order. Sort according Image Position along z
        slices = sorted(noDuplicateSlices, key=lambda s: s[0x0020, 0x0032][2])
    else:
        logger.error('no file available')
        return

    if len(slices) == 0:
        logger.error('no slice available')
        return

    dicomProperties = dicom_properties()
    if len(slices) >= 2:
        dicomProperties.read_dicom_properties(slices[0], slices[1])
    else:
        dicomProperties.read_dicom_properties(slices[0])

    # create 3D array
    dicomProperties.img_shape[0] = len(slices)
    dicomProperties.img_shape.append(slices[0].pixel_array.shape[0])
    img3d = np.zeros(dicomProperties.img_shape)

    # fill 3D array with the images from the files
    for i, s in enumerate(slices):
        img2d = s.pixel_array
        dicomPropertiesSlice = dicom_properties()
        dicomPropertiesSlice.read_dicom_slop_intercept(s)
        img3d[i, :, :] = dicomPropertiesSlice.rs*img2d+dicomPropertiesSlice.ri

    img_result = itk.image_from_array(np.float32(img3d))
    img_result.SetSpacing(dicomProperties.spacing)
    img_result.SetOrigin(dicomProperties.origin)
    arrayDirection = np.zeros([3,3], np.float64)
    arrayDirection[0,0] = dicomProperties.io[0]
    arrayDirection[0,1] = dicomProperties.io[1]
    arrayDirection[0,2] = dicomProperties.io[2]
    arrayDirection[1,0] = dicomProperties.io[3]
    arrayDirection[1,1] = dicomProperties.io[4]
    arrayDirection[1,2] = dicomProperties.io[5]
    arrayDirection[2,:] = np.cross(arrayDirection[0,:], arrayDirection[1,:])
    matrixItk = itk.Matrix[itk.D,3,3](itk.GetVnlMatrixFromArray(arrayDirection))
    img_result.SetDirection(matrixItk)
    return img_result


def read_3d_dicom(dicomFile, flip=False):
    """

    Read dicom file and return a float 3D image
    """
    files = []
    try:
        files.append(pydicom.read_file(dicomFile[0]))
    except pydicom.errors.InvalidDicomError:
        ds = pydicom.read_file(dicomFile[0], force=True)
        ds.file_meta.TransferSyntaxUID = pydicom.uid.ImplicitVRLittleEndian
        files.append(ds)

    slices = []
    if len(files) == 1:
        slices.append(files[0])
    else:
        logger.error('no file available')
        return

    dicomProperties = dicom_properties()
    dicomProperties.read_dicom_properties(slices[0])

    # create 3D array
    if len(slices[0].pixel_array.shape) == 2:
        dicomProperties.img_shape = [1] + dicomProperties.img_shape
    img3d = np.zeros(dicomProperties.img_shape)

    # fill 3D array with the images from the files
    if len(slices[0].pixel_array.shape) == 2:
        img3d[0, :, :] = slices[0].pixel_array
    else:
        img3d[:, :, :] = slices[0].pixel_array
    img3d = dicomProperties.rs*img3d+dicomProperties.ri

    img_result = itk.image_from_array(np.float32(img3d))
    img_result.SetSpacing(dicomProperties.spacing)
    img_result.SetOrigin(dicomProperties.origin)
    arrayDirection = np.zeros([3,3], np.float64)
    arrayDirection[0,0] = dicomProperties.io[0]
    arrayDirection[0,1] = dicomProperties.io[1]
    arrayDirection[0,2] = dicomProperties.io[2]
    arrayDirection[1,0] = dicomProperties.io[3]
    arrayDirection[1,1] = dicomProperties.io[4]
    arrayDirection[1,2] = dicomProperties.io[5]
    arrayDirection[2,:] = np.cross(arrayDirection[0,:], arrayDirection[1,:])
    if Tag(0x18, 0x88) in slices[0] and slices[0][0x18, 0x88].value <0:
        arrayDirection[2,2] = -1.0
    else:
        flip = False
        arrayDirection[2,2] = 1.0
    matrixItk = itk.Matrix[itk.D,3,3](itk.GetVnlMatrixFromArray(arrayDirection))
    img_result.SetDirection(matrixItk)
    if flip:
        flipFilter = itk.FlipImageFilter.New(Input=img_result)
        flipFilter.SetFlipAxes((False, False, True))
        flipFilter.SetFlipAboutOrigin(False)
        flipFilter.Update()
        img_result = flipFilter.GetOutput()
    return img_result

def image_convert(inputImage, pixeltype=None):
    """
    Convert the pixelType of the image

    - inputImage: input itk image to convert
    - pixeltype: string representing the type of the output pixel:
        - unsigned char
        - signed short
        - unsigned short
        - float
    """

    #If pixel type is not None, convert it,
    #If None it could be a type conversion without pixel conversion
    if pixeltype is not None:
        InputType = type(inputImage)
        OutputType = itk.Image[itk.ctype(pixeltype), inputImage.GetImageDimension()]
        castFilter = itk.CastImageFilter[InputType, OutputType].New()
        castFilter.SetInput(inputImage)
        castFilter.Update()
        return castFilter.GetOutput()

    return inputImage

#####################################################################################
import unittest
import hashlib
import os
import tempfile
import shutil
import wget
from .logging_conf import LoggedTestCase

def createImage():
    x = np.arange(-10, 10, 1)
    y = np.arange(-12, 15, 1)
    z = np.arange(-13, 10, 1)
    xx, yy, zz = np.meshgrid(x, y, z)
    return xx

class Test_Convert(LoggedTestCase):
    def test_convert_unsigned_char(self):
        image = itk.image_from_array(np.float32(createImage()))
        convertedImage = image_convert(image, "unsigned char")
        tmpdirpath = tempfile.mkdtemp()
        itk.imwrite(convertedImage, os.path.join(tmpdirpath, "testConvert.mha"))
        with open(os.path.join(tmpdirpath, "testConvert.mha"),"rb") as fnew:
            bytesNew = fnew.read()
            new_hash = hashlib.sha256(bytesNew).hexdigest()
            self.assertTrue("da57bf05ec62b9b5ee2aaa71aacbf7dbca8acbf2278553edc499a3af8007dd44" == new_hash)
        shutil.rmtree(tmpdirpath)
    def test_convert_short(self):
        image = itk.image_from_array(np.float32(createImage()))
        convertedImage = image_convert(image, "short")
        tmpdirpath = tempfile.mkdtemp()
        itk.imwrite(convertedImage, os.path.join(tmpdirpath, "testConvert.mha"))
        with open(os.path.join(tmpdirpath, "testConvert.mha"),"rb") as fnew:
            bytesNew = fnew.read()
            new_hash = hashlib.sha256(bytesNew).hexdigest()
            self.assertTrue("e71240bd149a509a3c28193817c82bb78d4cf5a6c8978b8266f820aa332cda0d" == new_hash)
        shutil.rmtree(tmpdirpath)
    def test_convert_unsigned_short(self):
        image = itk.image_from_array(np.float32(createImage()))
        convertedImage = image_convert(image, "unsigned short")
        tmpdirpath = tempfile.mkdtemp()
        itk.imwrite(convertedImage, os.path.join(tmpdirpath, "testConvert.mha"))
        with open(os.path.join(tmpdirpath, "testConvert.mha"),"rb") as fnew:
            bytesNew = fnew.read()
            new_hash = hashlib.sha256(bytesNew).hexdigest()
            self.assertTrue("55913ce0ca7d1c6f8166f0fdf7499e986d5965249742098cc2aaefe35964a1cb" == new_hash)
        shutil.rmtree(tmpdirpath)
    def test_convert_float(self):
        image = itk.image_from_array(np.int16(createImage()))
        convertedImage = image_convert(image, "float")
        tmpdirpath = tempfile.mkdtemp()
        itk.imwrite(convertedImage, os.path.join(tmpdirpath, "testConvert.mha"))
        with open(os.path.join(tmpdirpath, "testConvert.mha"),"rb") as fnew:
            bytesNew = fnew.read()
            new_hash = hashlib.sha256(bytesNew).hexdigest()
            self.assertTrue("c78c8b7450119dd554c338d6fd9f25648c5ee2863d8267cdd72dd20952673865" == new_hash)
        shutil.rmtree(tmpdirpath)
    def test_convert_rtDose(self):
        tmpdirpath = tempfile.mkdtemp()
        filenameRTDose = wget.download("https://gitlab.in2p3.fr/opengate/gatetools_data/-/raw/master/rtdose.dcm?inline=false", out=tmpdirpath, bar=None)
        convertedImage = read_3d_dicom([os.path.join(tmpdirpath, filenameRTDose)])
        itk.imwrite(convertedImage, os.path.join(tmpdirpath, "testConvert.mha"))
        with open(os.path.join(tmpdirpath, "testConvert.mha"),"rb") as fnew:
            bytesNew = fnew.read()
            new_hash = hashlib.sha256(bytesNew).hexdigest()
            self.assertTrue("9c18c0344e309d096122c4b771fe3f1e66ddb778d818f98d18f29442fd287d47" == new_hash)
        shutil.rmtree(tmpdirpath)
    def test_convert_dicom(self):
        tmpdirpath = tempfile.mkdtemp()
        filename = wget.download("https://gitlab.in2p3.fr/opengate/gatetools_data/-/raw/master/tomoSPECT.dcm?inline=false", out=tmpdirpath, bar=None)
        convertedImage = read_3d_dicom([os.path.join(tmpdirpath, filename)])
        itk.imwrite(convertedImage, os.path.join(tmpdirpath, "testConvert.mha"))
        with open(os.path.join(tmpdirpath, "testConvert.mha"),"rb") as fnew:
            bytesNew = fnew.read()
            new_hash = hashlib.sha256(bytesNew).hexdigest()
            self.assertTrue("3a64eee5f6439388fdbe6a5b7a682aca200fbc6826dcfffccc6f5fbae82b9600" == new_hash)
        shutil.rmtree(tmpdirpath)
    def test_convert_dicom_flip(self):
        tmpdirpath = tempfile.mkdtemp()
        filename = wget.download("https://gitlab.in2p3.fr/opengate/gatetools_data/-/raw/master/tomoSPECT.dcm?inline=false", out=tmpdirpath, bar=None)
        convertedImage = read_3d_dicom([os.path.join(tmpdirpath, filename)], flip=True)
        itk.imwrite(convertedImage, os.path.join(tmpdirpath, "testConvert.mha"))
        with open(os.path.join(tmpdirpath, "testConvert.mha"),"rb") as fnew:
            bytesNew = fnew.read()
            new_hash = hashlib.sha256(bytesNew).hexdigest()
            self.assertTrue("87c7eee6e29172289407e2739c2618418a38718c09e94ebee9a390a73433d236" == new_hash)
        shutil.rmtree(tmpdirpath)

